"""Text manipulation."""

import tensorflow as tf


def tokens_to_chars(tokens):
    """Splits tokens into unicode characters.

    Example:

      >>> opennmt.data.tokens_to_chars(["hello", "world"])
      <tf.RaggedTensor [[b'h', b'e', b'l', b'l', b'o'], [b'w', b'o', b'r', b'l', b'd']]>

    Args:
      tokens: A string ``tf.Tensor`` of shape :math:`[T]`.

    Returns:
      The characters as a 2D string ``tf.RaggedTensor``.
    """
    return tf.strings.unicode_split(tokens, "UTF-8")


def tokens_to_words(tokens, subword_token="￭", is_spacer=None):
    """Converts a sequence of tokens to a sequence of words.

    Example:

      >>> opennmt.data.tokens_to_words(["He@@", "llo", "W@@", "orld", "@@!"], subword_token="@@")
      <tf.RaggedTensor [[b'He@@', b'llo'], [b'W@@', b'orld', b'@@!']]>

    Args:
      tokens: A 1D string ``tf.Tensor``.
      subword_token: The special token used by the subword tokenizer.
      is_spacer: Whether :obj:`subword_token` is used as a spacer (as in
        SentencePiece) or a joiner (as in BPE). If ``None``, will infer
        directly from :obj:`subword_token`.

    Returns:
      The words as a 2D string ``tf.RaggedTensor``.
    """
    if is_spacer is None:
        is_spacer = subword_token == "▁"
    if is_spacer:
        # First token implicitly starts with a spacer.
        left_and_single = tf.logical_or(
            tf.strings.regex_full_match(tokens, "%s.*" % subword_token),
            tf.one_hot(0, tf.shape(tokens)[0], on_value=True, off_value=False),
        )
        right = tf.strings.regex_full_match(tokens, ".+%s" % subword_token)
        word_start = tf.logical_or(tf.roll(right, shift=1, axis=0), left_and_single)
    else:
        right = tf.strings.regex_full_match(tokens, ".*%s" % subword_token)
        left = tf.strings.regex_full_match(tokens, "%s.*" % subword_token)
        subword = tf.logical_or(tf.roll(right, shift=1, axis=0), left)
        word_start = tf.logical_not(subword)
    start_indices = tf.squeeze(tf.where(word_start), -1)
    return tf.RaggedTensor.from_row_starts(tokens, start_indices)


def alignment_matrix_from_pharaoh(
    alignment_line, source_length, target_length, dtype=tf.float32
):
    """Parse Pharaoh alignments into an alignment matrix.

    Example:

      >>> opennmt.data.alignment_matrix_from_pharaoh("0-0 1-2 1-3 2-1", 3, 4)
      <tf.Tensor: shape=(4, 3), dtype=float32, numpy=
      array([[1., 0., 0.],
             [0., 0., 1.],
             [0., 1., 0.],
             [0., 1., 0.]], dtype=float32)>

    Args:
      alignment_line: A string ``tf.Tensor`` in the Pharaoh format.
      source_length: The length of the source sentence, without special symbols.
      target_length: The length of the target sentence, without special symbols.
      dtype: The output matrix dtype. Defaults to ``tf.float32`` for convenience
        when computing the guided alignment loss.

    Returns:
      The alignment matrix as a 2-D ``tf.Tensor`` of type :obj:`dtype` and shape
      ``[target_length, source_length]``, where ``[i, j] = 1`` if the ``i`` th
      target word is aligned with the ``j`` th source word.
    """
    align_pairs_str = tf.strings.split([alignment_line]).values
    align_pairs_flat_str = tf.strings.split(align_pairs_str, sep="-").values
    align_pairs_flat = tf.strings.to_number(align_pairs_flat_str, out_type=tf.int64)
    sparse_indices = tf.reshape(align_pairs_flat, [-1, 2])
    sparse_values = tf.ones([tf.shape(sparse_indices)[0]], dtype=dtype)
    source_length = tf.cast(source_length, tf.int64)
    target_length = tf.cast(target_length, tf.int64)
    maximum_ids = tf.reduce_max(sparse_indices, axis=0)
    assert_source_length = _assert_in_range(
        maximum_ids[0], source_length, alignment_line, "source"
    )
    assert_target_length = _assert_in_range(
        maximum_ids[1], target_length, alignment_line, "target"
    )
    with tf.control_dependencies([assert_source_length, assert_target_length]):
        alignment_matrix_sparse = tf.sparse.SparseTensor(
            sparse_indices, sparse_values, [source_length, target_length]
        )
        alignment_matrix = tf.sparse.to_dense(
            alignment_matrix_sparse, validate_indices=False
        )
        return tf.transpose(alignment_matrix)


def _assert_in_range(maximum_id, length, line, name):
    return tf.debugging.assert_less(
        maximum_id,
        length,
        message=tf.strings.format(
            "Length mismatch for alignment line {}: actual %s length is {}, but "
            "got %s id {} which is out of range. Please check that the alignment "
            "file is correctly aligned to the training file." % (name, name),
            [line, length, maximum_id],
        ),
    )



"""Defines common layers."""

import tensorflow as tf

from opennmt.utils.misc import shape_list


def dropout(x, rate, training=None):
    """Simple dropout layer."""
    if not training or rate == 0:
        return x
    return tf.nn.dropout(x, rate)


def gelu(x):
    """Gaussian Error Linear Unit activation function described in
    https://arxiv.org/abs/1606.08415.
    """
    return tf.nn.gelu(x, approximate=True)


class Dense(tf.keras.layers.Dense):
    """Small ``tf.keras.layers.Dense`` extension to possibly reuse an existing weight
    matrix.
    """

    def __init__(self, units, weight=None, transpose=False, **kwargs):
        """Initializes the layer.

        Args:
          unit: Positive integer, dimensionality of the output space.
          weight: The weight to reuse.
          transpose: Whether :obj:`weight` should be transposed or not.
          kwargs: Additional layers arguments.
        """
        super().__init__(units, **kwargs)
        self.set_kernel(weight, transpose=transpose)

    def set_kernel(self, weight, transpose=False):
        """Use :obj:`weight` as the kernel weights matrix.

        Args:
          weight: The weight to use.
          transpose: Whether :obj:`weight` should be transposed or not.

        Raises:
          ValueError: if the layer is already built.
        """
        if self.built:
            raise ValueError("The layer is already built")
        self.weight = weight
        self.transpose = transpose

    def add_weight(self, name, *args, **kwargs):
        if self.weight is not None and name == "kernel":
            return self.weight
        return super().add_weight(name, *args, **kwargs)

    def call(self, inputs):
        shape = shape_list(inputs)
        rank = len(shape)
        if rank > 2:
            inputs = tf.reshape(inputs, [-1, shape[-1]])
        if inputs.dtype is tf.float16 and self.units % 8 != 0:
            padding_size = 8 - self.units % 8
            paddings = (
                [[0, padding_size], [0, 0]]
                if self.transpose
                else [[0, 0], [0, padding_size]]
            )
            kernel = tf.pad(self.kernel, paddings)
            outputs = tf.matmul(inputs, kernel, transpose_b=self.transpose)
            outputs = outputs[:, : self.units]
        else:
            outputs = tf.matmul(inputs, self.kernel, transpose_b=self.transpose)
        if self.use_bias:
            outputs = tf.nn.bias_add(outputs, self.bias)
        if self.activation is not None:
            outputs = self.activation(outputs)
        if rank > 2:
            outputs = tf.reshape(outputs, shape[:-1] + [self.units])
        return outputs

    def map_v1_weights(self, weights):
        m = [(self.kernel, weights["kernel"])]
        if self.use_bias:
            m.append((self.bias, weights["bias"]))
        return m


class LayerNorm(tf.keras.layers.LayerNormalization):
    """Layer normalization."""

    def map_v1_weights(self, weights):
        return [(self.beta, weights["beta"]), (self.gamma, weights["gamma"])]


class LayerWrapper(tf.keras.layers.Layer):
    """Layer wrapper for input/output normalization, input/output dropout and
    residual connection.
    """

    def __init__(
        self,
        layer,
        normalize_input=False,
        normalize_output=False,
        input_dropout=0,
        output_dropout=0,
        residual_connection=False,
        **kwargs
    ):
        """Initializes the layer.

        Args:
          layer: The layer to wrap.
          normalize_input: Apply layer normalization on the input.
          normalize_output: Apply layer normalization on the output.
          input_dropout: The probability to drop units in the layer input.
          output_dropout: The probability to drop units in the layer output.
          residual_connection: Add the inputs to layer outputs (if their shape are
            compatible).
          kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        self.layer = layer
        self.input_layer_norm = LayerNorm() if normalize_input else None
        self.output_layer_norm = LayerNorm() if normalize_output else None
        self.input_dropout = input_dropout
        self.output_dropout = output_dropout
        self.residual_connection = residual_connection

    def call(self, inputs, *args, **kwargs):
        """Runs the wrapper."""
        training = kwargs.get("training")
        x = inputs
        if self.input_layer_norm is not None:
            x = self.input_layer_norm(x)
        x = dropout(x, self.input_dropout, training=training)

        all_outputs = self.layer(x, *args, **kwargs)
        if isinstance(all_outputs, tuple):
            outputs = all_outputs[0]
            extra_outputs = list(all_outputs)[1:]
        else:
            outputs = all_outputs
            extra_outputs = None

        outputs = dropout(outputs, self.output_dropout, training=training)
        if self.residual_connection and outputs.shape[-1] == inputs.shape[-1]:
            outputs += inputs
        if self.output_layer_norm is not None:
            outputs = self.output_layer_norm(outputs)

        if extra_outputs:
            return tuple([outputs] + extra_outputs)
        return outputs

    # The wrapper should be serializable to be used in tf.keras.layers.Bidirectional.

    def get_config(self):
        """Returns the layer wrapper configuration."""
        config = {
            "layer": tf.keras.layers.serialize(self.layer),
            "normalize_input": self.input_layer_norm is not None,
            "normalize_output": self.output_layer_norm is not None,
            "input_dropout": self.input_dropout,
            "output_dropout": self.output_dropout,
            "residual_connection": self.residual_connection,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))

    @classmethod
    def from_config(cls, config):
        """Creates a layer wrapper from its configuration."""
        layer = tf.keras.layers.deserialize(config.pop("layer"))
        return cls(layer, **config)



"""RNN functions and classes for TensorFlow 2.0."""

import numpy as np
import tensorflow as tf

from opennmt.layers import common
from opennmt.layers import reducer as reducer_lib


class RNNCellWrapper(common.LayerWrapper):
    """A wrapper for RNN cells."""

    def __init__(
        self,
        cell,
        input_dropout=0,
        output_dropout=0,
        residual_connection=False,
        **kwargs
    ):
        """Initializes the wrapper.

        Args:
          cell: The cell to wrap.
          input_dropout: The probability to drop units in the cell input.
          output_dropout: The probability to drop units in the cell output.
          residual_connection: Add the inputs to cell outputs (if their shape are
            compatible).
          kwargs: Additional layer arguments.
        """
        super().__init__(
            cell,
            input_dropout=input_dropout,
            output_dropout=output_dropout,
            residual_connection=residual_connection,
            **kwargs,
        )
        self.cell = cell

    @property
    def state_size(self):
        """The cell state size."""
        return self.cell.state_size

    @property
    def output_size(self):
        """The cell output size."""
        return self.cell.output_size

    def get_initial_state(self, inputs=None, batch_size=None, dtype=None):
        """Returns the initial cell state."""
        return self.cell.get_initial_state(
            inputs=inputs, batch_size=batch_size, dtype=dtype
        )


def make_rnn_cell(
    num_layers,
    num_units,
    dropout=0,
    residual_connections=False,
    cell_class=None,
    **kwargs
):
    """Convenience function to build a multi-layer RNN cell.

    Args:
      num_layers: The number of layers.
      num_units: The number of units in each layer.
      dropout: The probability to drop units in each layer output.
      residual_connections: If ``True``, each layer input will be added to its output.
      cell_class: The inner cell class or a callable taking :obj:`num_units` as
        argument and returning a cell. Defaults to a LSTM cell.
      kwargs: Additional arguments passed to the cell constructor.

    Returns:
      A ``tf.keras.layers.StackedRNNCells`` instance.

    See Also:
      :class:`opennmt.layers.RNNCellWrapper`
    """
    if cell_class is None:
        cell_class = tf.keras.layers.LSTMCell
    cells = []
    for _ in range(num_layers):
        cell = cell_class(num_units, **kwargs)
        if dropout > 0 or residual_connections:
            cell = RNNCellWrapper(
                cell, output_dropout=dropout, residual_connection=residual_connections
            )
        cells.append(cell)
    return tf.keras.layers.StackedRNNCells(cells)


class _RNNWrapper(tf.keras.layers.Layer):
    """Extend a RNN layer to possibly make it bidirectional and format its outputs."""

    def __init__(
        self, rnn, bidirectional=False, reducer=reducer_lib.ConcatReducer(), **kwargs
    ):
        """Initializes the layer.

        Args:
          rnn: The RNN layer to extend, built with ``return_sequences`` and
            ``return_state`` enabled.
          bidirectional: Make this layer bidirectional.
          reducer: A :class:`opennmt.layers.Reducer` instance to merge
            bidirectional states and outputs.
          **kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        self.rnn = rnn
        self.reducer = reducer
        self.bidirectional = bidirectional
        if bidirectional:
            self.rnn = tf.keras.layers.Bidirectional(self.rnn, merge_mode=None)

    def call(self, *args, **kwargs):
        """Forwards the arguments to the RNN layer.

        Args:
          *args: Positional arguments of the RNN layer.
          **kwargs: Keyword arguments of the RNN layer.

        Returns:
          A tuple with the output sequences and the states.
        """
        outputs = self.rnn(*args, **kwargs)
        if self.bidirectional:
            sequences = outputs[0:2]
            states = outputs[2:]
            fwd_states = states[: len(states) // 2]
            bwd_states = states[len(states) // 2 :]
            if self.reducer is not None:
                sequences = self.reducer(sequences)
                states = tuple(self.reducer.zip_and_reduce(fwd_states, bwd_states))
            else:
                sequences = tuple(sequences)
                states = (fwd_states, bwd_states)
        else:
            sequences = outputs[0]
            states = tuple(outputs[1:])
        return sequences, states


class RNN(_RNNWrapper):
    """A simple RNN layer."""

    def __init__(
        self, cell, bidirectional=False, reducer=reducer_lib.ConcatReducer(), **kwargs
    ):
        """Initializes the layer.

        Args:
          cell: The RNN cell to use.
          bidirectional: Make this layer bidirectional.
          reducer: A :class:`opennmt.layers.Reducer` instance to merge
            bidirectional states and outputs.
          **kwargs: Additional layer arguments.

        See Also:
          :func:`opennmt.layers.make_rnn_cell`
        """
        rnn = tf.keras.layers.RNN(cell, return_sequences=True, return_state=True)
        super().__init__(rnn, bidirectional=bidirectional, reducer=reducer, **kwargs)

    def map_v1_weights(self, weights):
        m = []
        if self.bidirectional:
            weights = weights["bidirectional_rnn"]
            m += map_v1_weights_to_cell(self.rnn.forward_layer.cell, weights["fw"])
            m += map_v1_weights_to_cell(self.rnn.backward_layer.cell, weights["bw"])
        else:
            weights = weights["rnn"]
            m += map_v1_weights_to_cell(self.rnn.cell, weights)
        return m


class LSTM(tf.keras.layers.Layer):
    """A multi-layer LSTM.

    This differs from using :class:`opennmt.layers.RNN` with a ``LSTMCell`` in 2
    ways:

    - It uses ``tf.keras.layers.LSTM`` which is possibly accelerated by cuDNN on
      GPU.
    - Bidirectional outputs of each layer are reduced before feeding them to the
      next layer.
    """

    def __init__(
        self,
        num_layers,
        num_units,
        bidirectional=False,
        reducer=reducer_lib.ConcatReducer(),
        dropout=0,
        residual_connections=False,
        **kwargs
    ):
        """Initializes the layer.

        Args:
          num_layers: Number of stacked LSTM layers.
          num_units: Dimension of the output space of each LSTM.
          bidirectional: Make each layer bidirectional.
          reducer: A :class:`opennmt.layers.Reducer` instance to merge
            the bidirectional states and outputs of each layer.
          dropout: The probability to drop units in each layer output.
          residual_connections: If ``True``, each layer input will be added to its
            output.
          **kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        rnn_layers = [
            _RNNWrapper(
                tf.keras.layers.LSTM(
                    num_units, return_sequences=True, return_state=True
                ),
                bidirectional=bidirectional,
                reducer=reducer,
            )
            for _ in range(num_layers)
        ]
        self.layers = [
            common.LayerWrapper(
                layer, output_dropout=dropout, residual_connection=residual_connections
            )
            for layer in rnn_layers
        ]

    def call(self, inputs, mask=None, training=None, initial_state=None):
        all_states = []
        for i, layer in enumerate(self.layers):
            outputs, states = layer(
                inputs,
                mask=mask,
                training=training,
                initial_state=initial_state[i] if initial_state is not None else None,
            )
            all_states.append(states)
            inputs = outputs
        return outputs, tuple(all_states)


def map_v1_weights_to_cell(cell, weights):
    """Maps V1 weights to V2 RNN cell."""
    if isinstance(cell, RNNCellWrapper):
        cell = cell.cell
    if isinstance(cell, tf.keras.layers.StackedRNNCells):
        return _map_v1_weights_to_stacked_cells(cell, weights)
    elif isinstance(
        cell, (tf.keras.layers.LSTMCell, tf.compat.v1.keras.layers.LSTMCell)
    ):
        return _map_v1_weights_to_lstmcell(cell, weights)
    else:
        raise ValueError("Cannot restore V1 weights for cell %s" % str(cell))


def _map_v1_weights_to_stacked_cells(stacked_cells, weights):
    weights = weights["multi_rnn_cell"]
    m = []
    for i, cell in enumerate(stacked_cells.cells):
        m += map_v1_weights_to_cell(cell, weights["cell_%d" % i])
    return m


def _map_v1_weights_to_lstmcell(cell, weights):
    weights = weights["lstm_cell"]

    def _upgrade_weight(weight):
        is_bias = len(weight.shape) == 1
        i, j, f, o = np.split(weight, 4, axis=-1)
        if (
            is_bias
        ):  # Add forget_bias which is part of the LSTM formula in TensorFlow 1.
            f += 1
        return np.concatenate((i, f, j, o), axis=-1)  # Swap 2nd and 3rd projection.

    def _split_kernel(index):
        # TensorFlow 1 had a single kernel of shape [input_dim + units, 4 * units],
        # but TensorFlow 2 splits it into "kernel" and "recurrent_kernel".
        return tf.nest.map_structure(
            lambda w: np.split(w, [w.shape[0] - cell.units])[index], weights["kernel"]
        )

    weights = tf.nest.map_structure(_upgrade_weight, weights)
    m = []
    m.append((cell.kernel, _split_kernel(0)))
    m.append((cell.recurrent_kernel, _split_kernel(1)))
    if cell.use_bias:
        m.append((cell.bias, weights["bias"]))
    return m



"""Hypotheses file scoring for Precision Recall and F-Measure."""


def fmeasure(
    ref_path,
    hyp_path,
    return_precision_only=False,
    return_recall_only=False,
    return_fmeasure_only=False,
):
    """Compute Precision Recall and F-Measure between two files"""
    with open(ref_path) as ref_fp, open(hyp_path) as hyp_fp:
        list_null_tags = ["X", "null", "NULL", "Null", "O"]
        listtags = []
        classref = []
        classrandom = []
        classhyp = []
        nbrtagref = {}
        nbrtaghyp = {}
        nbrtagok = {}
        for line in ref_fp:
            line = line.strip()
            tabline = line.split(" ")
            lineref = []
            for tag in tabline:
                lineref.append(tag)
                if tag in nbrtagref.keys() and tag not in list_null_tags:
                    nbrtagref[tag] = nbrtagref[tag] + 1
                else:
                    nbrtagref[tag] = 1
            classref.append(lineref)
        for line, lineref in zip(hyp_fp, classref):
            line = line.strip()
            tabline = line.split(" ")
            linehyp = []
            linerandom = []
            for tagcpt, tag in enumerate(tabline):
                linehyp.append(tag)
                if tag not in listtags:
                    listtags.append(tag)
                linerandom.append(tag)
                if tagcpt < len(lineref) and tag == lineref[tagcpt]:
                    if tag in nbrtagok.keys():
                        nbrtagok[tag] = nbrtagok[tag] + 1
                    else:
                        nbrtagok[tag] = 1
                if tag in nbrtaghyp.keys():
                    nbrtaghyp[tag] = nbrtaghyp[tag] + 1
                else:
                    nbrtaghyp[tag] = 1
            classhyp.append(linehyp)
            classrandom.append(linerandom)

    fullprecision = 0
    fullrecall = 0
    precision = {}
    recall = {}
    fulltagok = 0.00
    fulltaghyp = 0.00
    fulltagref = 0.00
    for tag in listtags:
        if tag not in nbrtagok:
            nbrtagok[tag] = 0
        if tag not in nbrtaghyp:
            nbrtaghyp[tag] = 0
        if tag not in nbrtagref:
            nbrtagref[tag] = 0
        if nbrtaghyp[tag] != 0:
            precision[tag] = nbrtagok[tag] / nbrtaghyp[tag]
        else:
            precision[tag] = 0
        if nbrtagref[tag] != 0:
            recall[tag] = nbrtagok[tag] / nbrtagref[tag]
        else:
            recall[tag] = 0
        if tag not in list_null_tags:
            fulltagok = fulltagok + nbrtagok[tag]
            fulltaghyp = fulltaghyp + nbrtaghyp[tag]
            fulltagref = fulltagref + nbrtagref[tag]
    fullprecision = fulltagok / fulltaghyp if fulltaghyp != 0 else 0
    fullrecall = fulltagok / fulltagref if fulltagref != 0 else 0
    fullfmeasure = (
        (2 * fullprecision * fullrecall) / (fullprecision + fullrecall)
        if (fullprecision + fullrecall) != 0
        else 0
    )
    if return_precision_only:
        return fullprecision
    if return_recall_only:
        return fullrecall
    if return_fmeasure_only:
        return fullfmeasure
    return fullprecision, fullrecall, fullfmeasure



"""Vocabulary utilities for Python scripts."""

import numpy as np
import tensorflow as tf

from opennmt import constants


class Vocab(object):
    """Vocabulary class.

    Example:

      >>> vocab = opennmt.data.Vocab.from_file("wmtende.vocab")
      >>> len(vocab)
      32000
      >>> "be" in vocab
      True
      >>> vocab.lookup("be")
      377
      >>> vocab.lookup(377)
      'be'
    """

    def __init__(self, special_tokens=None):
        """Initializes a vocabulary.

        Args:
          special_tokens: A list of special tokens (e.g. start of sentence).
        """
        self._token_to_id = {}
        self._id_to_token = []
        self._frequency = []

        if special_tokens is not None:
            for index, token in enumerate(special_tokens):
                self._token_to_id[token] = index
                self._id_to_token.insert(index, token)

                # Set a very high frequency to avoid special tokens to be pruned.
                # Note that Python sort functions are stable which means that
                # special tokens in pruned vocabularies will have the same index.
                self._frequency.insert(index, float("inf"))

    @classmethod
    def from_file(cls, path, file_format="default"):
        """Creates from a vocabulary file.

        Args:
          path: The path to the vocabulary file.
          file_format: Define the format of the vocabulary file. Can be: default,
            sentencepiece. "default" is simply one token per line.

        Raises:
          ValueError: if :obj:`file_format` is invalid.
        """
        vocab = cls()
        vocab.load(path, file_format=file_format)
        return vocab

    @property
    def size(self):
        """Returns the number of entries of the vocabulary."""
        return len(self._id_to_token)

    @property
    def words(self):
        """Returns the list of words."""
        return self._id_to_token

    def __len__(self):
        """Returns the number of entries of the vocabulary."""
        return self.size

    def __contains__(self, token):
        """Returns ``True`` if the vocabulary contains :obj:`token`."""
        return self.lookup(token) is not None

    def add_from_text(self, filename, tokenizer=None):
        """Fills the vocabulary from a text file.

        Args:
          filename: The file to load from.
          tokenizer: A callable to tokenize a line of text.
        """
        with tf.io.gfile.GFile(filename) as text:
            for line in text:
                line = line.rstrip("\r\n")
                if tokenizer:
                    tokens = tokenizer.tokenize(line)
                else:
                    tokens = line.split()
                for token in tokens:
                    self.add(token)

    def serialize(self, path):
        """Writes the vocabulary on disk.

        Args:
          path: The path where the vocabulary will be saved.
        """
        with tf.io.gfile.GFile(path, mode="w") as vocab:
            for token in self._id_to_token:
                vocab.write(token)
                vocab.write("\n")

    def load(self, path, file_format="default"):
        """Loads a serialized vocabulary.

        Args:
          path: The path to the vocabulary to load.
          file_format: Define the format of the vocabulary file. Can be: default,
            sentencepiece. "default" is simply one token per line.

        Raises:
          ValueError: if :obj:`file_format` is invalid.
        """
        if file_format not in ("default", "sentencepiece"):
            raise ValueError("Invalid vocabulary format: %s" % file_format)
        with tf.io.gfile.GFile(path) as vocab:
            for i, line in enumerate(vocab):
                token = line.rstrip("\r\n")
                if file_format == "sentencepiece":
                    token, _ = token.split("\t")
                    # Ignore SentencePiece special tokens.
                    if token in ("<unk>", "<s>", "</s>"):
                        continue

                if token in self._token_to_id:
                    tf.get_logger().warning(
                        "Duplicate token '%s' in vocabulary %s at line %d",
                        token,
                        path,
                        i + 1,
                    )
                    continue

                self._token_to_id[token] = len(self._id_to_token)
                self._id_to_token.append(token)
                self._frequency.append(1)

    def add(self, token):
        """Adds a token or increases its frequency.

        Args:
          token: The string to add.
        """
        token = tf.compat.as_text(token)
        if token not in self._token_to_id:
            index = self.size
            self._token_to_id[token] = index
            self._id_to_token.append(token)
            self._frequency.append(1)
        else:
            self._frequency[self._token_to_id[token]] += 1

    def lookup(self, identifier, default=None):
        """Lookups in the vocabulary.

        Args:
          identifier: A string or an index to lookup.
          default: The value to return if :obj:`identifier` is not found.

        Returns:
          The value associated with :obj:`identifier` or :obj:`default`.
        """
        value = None

        if isinstance(identifier, (bytes, str)):
            identifier = tf.compat.as_text(identifier)
            value = self._token_to_id.get(identifier)
        elif identifier < self.size:
            value = self._id_to_token[identifier]

        if value is None:
            return default
        else:
            return value

    def prune(self, max_size=0, min_frequency=1):
        """Creates a pruned version of the vocabulary.

        Args:
          max_size: The maximum vocabulary size.
          min_frequency: The minimum frequency of each entry.

        Returns:
          A new vocabulary.
        """
        sorted_ids = sorted(
            range(self.size), key=lambda k: self._frequency[k], reverse=True
        )
        new_size = len(sorted_ids)

        # Discard words that do not meet frequency requirements.
        for i in range(new_size - 1, 0, -1):
            index = sorted_ids[i]
            if self._frequency[index] < min_frequency:
                new_size -= 1
            else:
                break

        # Limit absolute size.
        if max_size > 0:
            new_size = min(new_size, max_size)

        new_vocab = Vocab()

        for i in range(new_size):
            index = sorted_ids[i]
            token = self._id_to_token[index]
            frequency = self._frequency[index]

            new_vocab._token_to_id[token] = i
            new_vocab._id_to_token.append(token)
            new_vocab._frequency.append(frequency)

        return new_vocab

    def pad_to_multiple(self, multiple, num_oov_buckets=1):
        """Pads the vocabulary size to a multiple value.

        More specically, this method ensures that:

            ``(vocab_size + num_oov_buckets) % multiple == 0``

        Args:
          multiple: The multiple value.
          num_oov_buckets: The number of OOV buckets added during the training.
            Usually just 1 for the `<unk>` token.
        """
        i = 0
        while (self.size + num_oov_buckets) % multiple != 0:
            self.add("averyunlikelytoken%d" % i)
            i += 1


def create_lookup_tables(
    vocabulary_path, num_oov_buckets=1, as_asset=True, unk_token=None
):
    """Creates TensorFlow lookup tables from a vocabulary file.

    Args:
      vocabulary_path: Path to the vocabulary file.
      num_oov_buckets: Number of out-of-vocabulary buckets.
      as_asset: If ``True``, the vocabulary file will be added as a graph asset.
        Otherwise, the content of the vocabulary will be embedded in the graph.
      unk_token: The out-of-vocabulary token. Defaults to ``<unk>``.

    Returns:
      A tuple containing,

      - The final vocabulary size.
      - The ``tf.lookup`` table mapping tokens to ids.
      - The ``tf.lookup`` table mapping ids to tokens.
    """
    if unk_token is None:
        unk_token = constants.UNKNOWN_TOKEN
    vocabulary = Vocab.from_file(vocabulary_path)
    vocabulary_size = len(vocabulary)
    if as_asset:
        tokens_to_ids_initializer = tf.lookup.TextFileInitializer(
            vocabulary_path,
            tf.string,
            tf.lookup.TextFileIndex.WHOLE_LINE,
            tf.int64,
            tf.lookup.TextFileIndex.LINE_NUMBER,
            vocab_size=vocabulary_size,
        )
        ids_to_tokens_initializer = tf.lookup.TextFileInitializer(
            vocabulary_path,
            tf.int64,
            tf.lookup.TextFileIndex.LINE_NUMBER,
            tf.string,
            tf.lookup.TextFileIndex.WHOLE_LINE,
            vocab_size=vocabulary_size,
        )
    else:
        tokens = tf.constant(vocabulary.words, dtype=tf.string)
        ids = tf.constant(list(range(vocabulary_size)), dtype=tf.int64)
        tokens_to_ids_initializer = tf.lookup.KeyValueTensorInitializer(tokens, ids)
        ids_to_tokens_initializer = tf.lookup.KeyValueTensorInitializer(ids, tokens)
    if num_oov_buckets > 0:
        tokens_to_ids = tf.lookup.StaticVocabularyTable(
            tokens_to_ids_initializer, num_oov_buckets
        )
    else:
        tokens_to_ids = tf.lookup.StaticHashTable(tokens_to_ids_initializer, 0)
    ids_to_tokens = tf.lookup.StaticHashTable(ids_to_tokens_initializer, unk_token)
    return vocabulary_size + num_oov_buckets, tokens_to_ids, ids_to_tokens


def get_mapping(current_vocab_path, new_vocab_path, mode="replace"):
    """Maps vocabulary new indices to old ones. -1 means that the entry is new."""
    mode = mode.lower()
    if mode not in ("merge", "replace"):
        raise ValueError("invalid vocab update mode: %s" % mode)
    current_vocab = Vocab.from_file(current_vocab_path)
    new_vocab = Vocab.from_file(new_vocab_path)
    mapping = []
    if mode == "merge":
        final_vocab = Vocab.from_file(current_vocab_path)
        mapping = list(range(current_vocab.size))
        for new_word in new_vocab.words:
            if current_vocab.lookup(new_word) is None:
                mapping.append(-1)
                final_vocab.add(new_word)
    elif mode == "replace":
        final_vocab = new_vocab
        for new_word in new_vocab.words:
            idx = current_vocab.lookup(new_word)
            if idx is not None:
                mapping.append(idx)
            else:
                mapping.append(-1)
    mapping.append(current_vocab.size)  # <unk> token is always the last entry.
    return mapping, final_vocab


def update_variable(ref_variable, new_variable, mapping, vocab_axis=0):
    """Update a vocabulary variable, possibly copying previous entries based on
    mapping.
    """
    ref = ref_variable.numpy()
    new = np.zeros(
        new_variable.shape.as_list(), dtype=new_variable.dtype.as_numpy_dtype
    )
    perm = None
    if vocab_axis != 0:
        # Make the dimension to index the first.
        perm = list(range(len(ref.shape)))
        perm[0], perm[vocab_axis] = perm[vocab_axis], perm[0]
        ref = np.transpose(ref, axes=perm)
        new = np.transpose(new, axes=perm)
    for i, j in enumerate(mapping):
        if j >= 0:
            new[i] = ref[j]
    if perm is not None:
        new = np.transpose(new, axes=perm)
    new_variable.assign(new)
    return new_variable


def update_variable_and_slots(
    ref_variable, new_variable, ref_optimizer, new_optimizer, mapping, vocab_axis=0
):
    """Update a vocabulary variable and its associated optimizer slots (if any)."""
    variables = []
    variables.append(
        update_variable(ref_variable, new_variable, mapping, vocab_axis=vocab_axis)
    )
    ref_slot_names = ref_optimizer.get_slot_names()
    new_slot_names = new_optimizer.get_slot_names()
    for slot_name in ref_slot_names:
        if slot_name not in new_slot_names:
            continue
        ref_slot = ref_optimizer.get_slot(ref_variable, slot_name)
        new_slot = new_optimizer.get_slot(new_variable, slot_name)
        variables.append(
            update_variable(ref_slot, new_slot, mapping, vocab_axis=vocab_axis)
        )
    return variables



"""Module defining various utilities."""

from opennmt.utils.checkpoint import (
    average_checkpoints,
    average_checkpoints_into_layer,
    is_v1_checkpoint,
)
from opennmt.utils.decoding import (
    BeamSearch,
    BestSampler,
    DecodingResult,
    DecodingStrategy,
    GreedySearch,
    RandomSampler,
    Sampler,
    dynamic_decode,
)
from opennmt.utils.exporters import (
    CheckpointExporter,
    CTranslate2Exporter,
    CTranslate2Float16Exporter,
    CTranslate2Int8Exporter,
    CTranslate2Int8Float16Exporter,
    CTranslate2Int16Exporter,
    Exporter,
    SavedModelExporter,
    TFLiteExporter,
    TFLiteFloat16Exporter,
    register_exporter,
)
from opennmt.utils.losses import (
    cross_entropy_loss,
    cross_entropy_sequence_loss,
    guided_alignment_cost,
    max_margin_loss,
    regularization_penalty,
)
from opennmt.utils.misc import format_translation_output
from opennmt.utils.scorers import (
    BLEUScorer,
    PRFScorer,
    ROUGEScorer,
    Scorer,
    TERScorer,
    WERScorer,
    make_scorers,
    register_scorer,
)
from opennmt.utils.tensor import roll_sequence



"""Module defining decoders."""

from opennmt.decoders.decoder import Decoder, get_sampling_probability
from opennmt.decoders.rnn_decoder import (
    AttentionalRNNDecoder,
    RNMTPlusDecoder,
    RNNDecoder,
)
from opennmt.decoders.self_attention_decoder import SelfAttentionDecoder



"""Standalone script to generate word vocabularies from monolingual corpus."""

import argparse

import tensorflow as tf

from opennmt import constants, data, tokenizers


def main():
    tf.get_logger().setLevel("INFO")

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument("data", nargs="*", help="List of data files.")
    parser.add_argument(
        "--from_vocab",
        default=None,
        help="Build from a saved vocabulary (see also --from_format).",
    )
    parser.add_argument(
        "--from_format",
        default="default",
        choices=["default", "sentencepiece"],
        help="The format of the saved vocabulary (see also --from_vocab).",
    )
    parser.add_argument("--save_vocab", required=True, help="Output vocabulary file.")
    parser.add_argument(
        "--min_frequency", type=int, default=1, help="Minimum word frequency."
    )
    parser.add_argument(
        "--size",
        type=int,
        default=0,
        help="Maximum vocabulary size. If = 0, do not limit vocabulary.",
    )
    parser.add_argument(
        "--size_multiple",
        type=int,
        default=1,
        help=(
            "Ensure that the vocabulary size + 1 is a multiple of this value "
            "(+ 1 represents the <unk> token that will be added during the training."
        ),
    )
    parser.add_argument(
        "--without_sequence_tokens",
        default=False,
        action="store_true",
        help="If set, do not add special sequence tokens (start, end) in the vocabulary.",
    )
    parser.add_argument(
        "--tokenizer_config",
        default=None,
        help=(
            "Tokenization configuration as a JSON string or a path to a YAML configuration file. "
            "When building a SentencePiece model and vocabulary, this is used as a "
            "pre-tokenization. SentencePiece will receive tokens instead of sentences as "
            "inputs."
        ),
    )
    parser.add_argument(
        "--sentencepiece",
        nargs="*",
        default=None,
        help=(
            "Build a SentencePiece model and vocabulary. This option accepts additional "
            "training parameters (e.g. --sentencepiece character_coverage=0.98)."
        ),
    )
    args = parser.parse_args()

    special_tokens = [constants.PADDING_TOKEN]
    if not args.without_sequence_tokens:
        special_tokens.append(constants.START_OF_SENTENCE_TOKEN)
        special_tokens.append(constants.END_OF_SENTENCE_TOKEN)

    vocab = data.Vocab(special_tokens=special_tokens)
    num_oov_buckets = 1

    if args.sentencepiece is not None:
        if args.min_frequency > 1:
            raise ValueError(
                "--min_frequency option is not supported when training a SentencePiece "
                "model and vocabulary"
            )

        import pyonmttok

        if args.size_multiple == 1:
            vocab_size = args.size
        else:
            # Round vocabulary size to the next multiple of args.size_multiple
            vocab_size = (
                args.size
                - (args.size + num_oov_buckets) % args.size_multiple
                + args.size_multiple
            )

        if args.tokenizer_config:
            tokenizer = tokenizers.make_tokenizer(args.tokenizer_config)
            if not isinstance(tokenizer, tokenizers.OpenNMTTokenizer):
                tokenizer_type = tokenizer.__class__.__name__
                raise ValueError(
                    "Only tokenizer type 'OpenNMTTokenizer' can be used as a SentencePiece "
                    "pre-tokenization, got tokenizer type '%s' instead."
                    % tokenizer_type
                )
        else:
            tokenizer = None

        sp_params = dict(map(lambda arg: tuple(arg.split("=")), args.sentencepiece))
        sp_trainer = pyonmttok.SentencePieceLearner(
            tokenizer=tokenizer.opennmt_tokenizer if tokenizer is not None else None,
            keep_vocab=True,
            vocab_size=vocab_size,
            **sp_params,
        )

        for data_file in args.data:
            sp_trainer.ingest_file(data_file)
        sp_trainer.learn(args.save_vocab, verbose=True)

        model_path = args.save_vocab + ".model"
        vocab_path = args.save_vocab + ".vocab"

        if tokenizer is None:
            tf.get_logger().info(
                "Converting SentencePiece vocabulary to OpenNMT-tf format..."
            )
            vocab.load(vocab_path, file_format="sentencepiece")
        else:
            tf.get_logger().info(
                "Applying SentencePiece model on data and extracting the %d most "
                "frequent tokens...",
                vocab_size,
            )
            tokenizer = tokenizers.OpenNMTTokenizer(
                sp_model_path=model_path, **tokenizer.config
            )
            for data_file in args.data:
                vocab.add_from_text(data_file, tokenizer=tokenizer)
            vocab = vocab.prune(max_size=vocab_size)

        vocab.serialize(vocab_path)

    else:
        if args.from_vocab is not None:
            vocab.load(args.from_vocab, file_format=args.from_format)
        tokenizer = tokenizers.make_tokenizer(args.tokenizer_config)
        for data_file in args.data:
            vocab.add_from_text(data_file, tokenizer=tokenizer)
        vocab = vocab.prune(max_size=args.size, min_frequency=args.min_frequency)
        vocab.pad_to_multiple(args.size_multiple, num_oov_buckets=num_oov_buckets)
        vocab.serialize(args.save_vocab)


if __name__ == "__main__":
    main()



"""Noise modules."""

import abc

import tensorflow as tf

from opennmt import constants
from opennmt.data import text
from opennmt.utils import misc


class WordNoiser(object):
    """Applies noise to words sequences."""

    def __init__(self, noises=None, subword_token="￭", is_spacer=None):
        """Initializes the noising class.

        Args:
          noises: A list of :class:`opennmt.data.Noise` instances to apply
            sequentially.
          subword_token: The special token used by the subword tokenizer. This is
            required when the noise should be applied at the word level and not the
            subword level.
          is_spacer: Whether :obj:`subword_token` is used as a spacer (as in
            SentencePiece) or a joiner (as in BPE). If ``None``, will infer
            directly from :obj:`subword_token`.

        See Also:
          :func:`opennmt.data.tokens_to_words`
        """
        if noises is None:
            noises = []
        self.noises = noises
        self.subword_token = subword_token
        self.is_spacer = is_spacer

    def add(self, noise):
        """Adds a noise to apply."""
        self.noises.append(noise)

    def __call__(
        self, tokens, sequence_length=None, keep_shape=False, probability=None
    ):
        """Applies noise on :obj:`tokens`.

        Args:
          tokens: A string ``tf.Tensor``, a batch of string ``tf.Tensor``, or a
            string ``tf.RaggedTensor``.
          sequence_length: When :obj:`tokens` is a dense tensor, the length of
            each sequence in the batch.
          keep_shape: Ensure that the original dense shape is kept. Otherwise,
            fit the shape to the new lengths.
          probability: Probability to apply noise on each example.

        Returns:
          If :obj:`tokens` is a ``tf.RaggedTensor``, the method returns the
          noisy tokens as a ``tf.RaggedTensor``, otherwise it returns a tuple
          with the noisy tokens as a ``tf.Tensor`` and the new lengths.

        Raises:
          ValueError: if :obj:`tokens` is a batch of string but
            :obj:`sequence_length` is not passed.
          ValueError: if :obj:`keep_shape` is ``True`` but :obj:`tokens` is a
            ``tf.RaggedTensor``.
        """
        if probability is None:
            probability = 1
        with tf.device("cpu:0"):
            return self._call(tokens, sequence_length, keep_shape, probability)

    def _call(self, tokens, sequence_length, keep_shape, probability):
        input_rank = tokens.shape.rank
        input_is_ragged = isinstance(tokens, tf.RaggedTensor)

        if input_is_ragged:
            if keep_shape:
                raise ValueError("keep_shape is not compatible with a ragged input")
            ragged_tokens = tokens
        elif input_rank == 1:
            tokens = tf.expand_dims(tokens, 0)
            ragged_tokens = tf.RaggedTensor.from_tensor(tokens)
        elif input_rank == 2:
            if sequence_length is None:
                raise ValueError("sequence_length must be passed for 2D dense inputs")
            ragged_tokens = tf.RaggedTensor.from_tensor(tokens, lengths=sequence_length)
        else:
            raise ValueError("unsupported rank %d for WordNoiser input" % input_rank)

        noisy_tokens = tf.map_fn(
            lambda tokens: self._maybe_apply_noise(tokens, probability),
            ragged_tokens,
            fn_output_signature=tf.RaggedTensorSpec(
                shape=[None], dtype=ragged_tokens.dtype, ragged_rank=0
            ),
        )

        if input_is_ragged:
            return noisy_tokens

        new_lengths = tf.cast(noisy_tokens.row_lengths(), tf.int32)
        noisy_tokens = noisy_tokens.to_tensor(
            shape=tf.shape(tokens) if keep_shape else None
        )
        if input_rank == 1:
            new_lengths = new_lengths[0]
            noisy_tokens = noisy_tokens[0]

        return noisy_tokens, new_lengths

    def _maybe_apply_noise(self, tokens, probability):
        if probability == 1:
            return self._apply_noise(tokens)
        elif probability == 0:
            return tokens
        else:
            return tf.cond(
                random_mask([], probability),
                true_fn=lambda: self._apply_noise(tokens),
                false_fn=lambda: tokens,
            )

    def _apply_noise(self, tokens):
        words = text.tokens_to_words(
            tokens, subword_token=self.subword_token, is_spacer=self.is_spacer
        )
        words = words.to_tensor()
        for noise in self.noises:
            words = noise(words)
        tokens = tf.RaggedTensor.from_tensor(words, padding="").flat_values
        return tokens


class Noise(abc.ABC):
    """Base class for noise modules."""

    def __call__(self, words):
        """Applies noise on a sequence of words.

        Args:
          words: The sequence of words as a string ``tf.Tensor``. If it has 2
            dimensions, each row represents a word that possibly contains multiple
            tokens.

        Returns:
          A noisy version of :obj:`words`.

        Raises:
          ValueError: if :obj:`words` has a rank greater than 2.
        """
        if words.shape.ndims > 2:
            raise ValueError("Noise only supports tensors of rank 2 or less")
        inputs = words
        if words.shape.ndims == 1:
            inputs = tf.expand_dims(inputs, 1)
        num_words = tf.shape(inputs)[0]
        outputs = tf.cond(
            tf.math.equal(num_words, 0),
            true_fn=lambda: inputs,
            false_fn=lambda: self._apply(inputs),
        )
        if words.shape.ndims == 1:
            outputs = tf.squeeze(outputs, 1)
        return outputs

    @abc.abstractmethod
    def _apply(self, words):
        """Applies noise on a sequence of words.

        Args:
          words: A 2D string ``tf.Tensor`` where each row represents a word that
            possibly contains multiple tokens.

        Returns:
          A noisy version of :obj:`words`.
        """
        raise NotImplementedError()


class WordDropout(Noise):
    """Randomly drops words in a sequence.

    Example:

      >>> noise = opennmt.data.WordDropout(0.5)
      >>> words = tf.constant(["a", "b", "c"])
      >>> noise(words).numpy()
      array([b'a', b'b'], dtype=object)
    """

    def __init__(self, dropout):
        """Initializes the noise module.

        Args:
          dropout: The probability to drop word.
        """
        self.dropout = dropout

    def _apply(self, words):
        if self.dropout == 0:
            return tf.identity(words)
        num_words = tf.shape(words, out_type=tf.int64)[0]
        keep_mask = random_mask([num_words], 1 - self.dropout)
        keep_ind = tf.where(keep_mask)
        # Keep at least one word.
        keep_ind = tf.cond(
            tf.equal(tf.shape(keep_ind)[0], 0),
            true_fn=lambda: tf.random.uniform([1], maxval=num_words, dtype=tf.int64),
            false_fn=lambda: tf.squeeze(keep_ind, -1),
        )
        return tf.gather(words, keep_ind)


class WordOmission(Noise):
    """Randomly omits words in a sequence.

    This is different than :class:`opennmt.data.WordDropout` as it drops a
    fixed number of words.

    Example:

      >>> noise = opennmt.data.WordOmission(1)
      >>> words = tf.constant(["a", "b", "c"])
      >>> noise(words).numpy()
      array([b'b', b'c'], dtype=object)
    """

    def __init__(self, count):
        """Initializes the noise module.

        Args:
          count: The number of words to omit.
        """
        self.count = count

    def _apply(self, words):
        if self.count == 0:
            return tf.identity(words)
        num_words = tf.shape(words)[0]
        indices = tf.range(num_words)
        shuffle_indices = tf.random.shuffle(indices)
        keep_count = tf.maximum(num_words - self.count, 1)
        keep_indices = tf.sort(shuffle_indices[:keep_count])
        return tf.gather(words, keep_indices)


class WordReplacement(Noise):
    """Randomly replaces words.

    Example:

      >>> noise = opennmt.data.WordReplacement(0.5)
      >>> words = tf.constant(["a", "b", "c"])
      >>> noise(words).numpy()
      array([b'a', b'<unk>', b'c'], dtype=object)
    """

    def __init__(self, probability, filler=constants.UNKNOWN_TOKEN):
        """Initializes the noise module.

        Args:
          probability: The probability to replace words.
          filler: The replacement token.
        """
        self.probability = probability
        self.filler = filler

    def _apply(self, words):
        if self.probability == 0:
            return tf.identity(words)
        shape = tf.shape(words)
        replace_mask = random_mask(shape[:1], self.probability)
        filler = tf.fill([shape[0], 1], self.filler)
        filler = tf.pad(filler, [[0, 0], [0, shape[-1] - 1]])
        return tf.where(
            tf.broadcast_to(tf.expand_dims(replace_mask, -1), tf.shape(words)),
            x=filler,
            y=words,
        )


class WordPermutation(Noise):
    """Randomly permutes words in a sequence with a maximum distance.

    Example:

      >>> noise = opennmt.data.WordPermutation(3)
      >>> words = tf.constant(["0", "1", "2", "3", "4", "5", "6"])
      >>> noise(words).numpy()
      array([b'1', b'0', b'2', b'4', b'3', b'6', b'5'], dtype=object)
    """

    def __init__(self, max_distance):
        """Initializes the noise module.

        Args:
          max_distance: The maximum permutation distance.
        """
        self.max_distance = max_distance

    def _apply(self, words):
        if self.max_distance == 0:
            return tf.identity(words)
        num_words = tf.shape(words)[0]
        offset = tf.random.uniform([num_words], maxval=1) * (self.max_distance + 1)
        offset = tf.cast(offset, num_words.dtype)
        new_pos = tf.argsort(tf.range(num_words) + offset)
        return tf.gather(words, new_pos)


def random_mask(shape, probability):
    """Generates a random boolean mask.

    Args:
      shape: The mask shape.
      probability: The probability to select an element.

    Returns:
      A boolean mask with shape :obj:`shape`.
    """
    probs = tf.random.uniform(shape, maxval=1)
    mask = tf.math.less(probs, probability)
    return mask



"""Define learning rate decay functions."""

import inspect

import numpy as np
import tensorflow as tf

from opennmt.utils import misc

_LR_SCHEDULES_REGISTRY = misc.ClassRegistry(
    base_class=tf.keras.optimizers.schedules.LearningRateSchedule
)

register_learning_rate_schedule = _LR_SCHEDULES_REGISTRY.register


def get_lr_schedule_class(name):
    """Returns the learning rate schedule class.

    Args:
      name: The schedule class name.

    Returns:
      A class extending ``tf.keras.optimizers.schedules.LearningRateSchedule``.

    Raises:
      ValueError: if :obj:`name` can not be resolved to an existing schedule.
    """
    schedule_class = None
    if schedule_class is None:
        schedule_class = getattr(tf.keras.optimizers.schedules, name, None)
    if schedule_class is None:
        schedule_class = _LR_SCHEDULES_REGISTRY.get(name)
    if schedule_class is None:
        raise ValueError("Unknown learning rate schedule: %s" % name)
    return schedule_class


def make_learning_rate_schedule(
    initial_learning_rate,
    schedule_type,
    schedule_params=None,
    schedule_step_duration=1,
    start_step=0,
    minimum_learning_rate=0,
):
    """Creates the learning rate schedule.

    Args:
      initial_learning_rate: The initial learning rate value. This can be
        ``None`` if the learning rate is fully defined by the schedule.
      schedule_type: The type of learning rate schedule. A class name from
        ``tf.keras.optimizers.schedules`` or :mod:`opennmt.schedules` as a string.
      schedule_params: Additional parameters passed to the schedule constructor.
      schedule_step_duration: The number of training steps that make 1 schedule step.
      start_step: Start the schedule after this many steps.
      minimum_learning_rate: Do not decay past this learning rate value.

    Returns:
      A ``tf.keras.optimizers.schedules.LearningRateSchedule`` instance.

    Raises:
      ValueError: if :obj:`schedule_type` can not be resolved to an existing
        schedule.

    See Also:
      :class:`opennmt.schedules.ScheduleWrapper`
    """
    if schedule_params is None:
        schedule_params = {}
    schedule_class = get_lr_schedule_class(schedule_type)
    first_arg = inspect.getfullargspec(schedule_class)[0][1]
    if first_arg not in schedule_params:
        schedule_params[first_arg] = initial_learning_rate
    schedule = schedule_class(**schedule_params)
    schedule = ScheduleWrapper(
        schedule,
        step_start=start_step,
        step_duration=schedule_step_duration,
        minimum_learning_rate=minimum_learning_rate,
    )
    return schedule


class ScheduleWrapper(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Wrapper to augment a learning rate scheduler behavior."""

    def __init__(
        self, schedule, step_start=0, step_duration=1, minimum_learning_rate=0
    ):
        """Initializes the decay function.

        Args:
          schedule: A ``tf.keras.optimizers.schedules.LearningRateSchedule``.
          step_duration: The number of training steps that make 1 decay step.
          start_step: Start decay after this many steps.
          minimum_learning_rate: Do not decay past this learning rate value.

        See Also:
          :class:`opennmt.schedules.make_learning_rate_schedule`
        """
        self.schedule = schedule
        self.step_start = step_start
        self.step_duration = step_duration
        self.minimum_learning_rate = minimum_learning_rate

    def __call__(self, step):
        # Map the training step to a decay step.
        step = tf.maximum(step - self.step_start, 0)
        step //= self.step_duration
        learning_rate = self.schedule(step)
        return tf.maximum(learning_rate, self.minimum_learning_rate)


@register_learning_rate_schedule
class NoamDecay(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Defines the decay function described in https://arxiv.org/abs/1706.03762."""

    def __init__(self, scale, model_dim, warmup_steps):
        """Initializes the decay function.

        Args:
          scale: The scale constant.
          model_dim: The model dimension.
          warmup_steps: The number of warmup steps.
        """
        self.scale = tf.cast(scale, tf.float32)
        self.model_dim = tf.cast(model_dim, tf.float32)
        self.warmup_steps = tf.cast(warmup_steps, tf.float32)

    def __call__(self, step):
        step = tf.cast(step + 1, tf.float32)
        return (
            self.scale
            * tf.pow(self.model_dim, -0.5)
            * tf.minimum(tf.pow(step, -0.5), step * tf.pow(self.warmup_steps, -1.5))
        )


@register_learning_rate_schedule
class RsqrtDecay(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Decay based on the reciprocal of the step square root."""

    def __init__(self, scale, warmup_steps):
        """Initializes the decay function.

        Args:
          scale: The scale constant.
          warmup_steps: The number of warmup steps.
        """
        self.scale = tf.cast(scale, tf.float32)
        self.warmup_steps = tf.cast(warmup_steps, tf.float32)

    def __call__(self, step):
        step = tf.cast(step, tf.float32)
        return self.scale * tf.math.rsqrt(tf.maximum(step, self.warmup_steps))


@register_learning_rate_schedule
class CosineAnnealing(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Decay using a cosine annealing schedule."""

    def __init__(self, eta_max, eta_min=0, max_step=1000000, warmup_steps=None):
        """Initializes the decay function.

        Args:
          eta_max: Maximum learning rate.
          eta_min: Minimum learning rate.
          max_step: The last step of the scedule.
          warmup_steps: The number of steps to increment the learning rate linearly
            from 0 to :obj:`scale` before annealing.
        """
        self.eta_max = tf.cast(eta_max, tf.float32)
        self.eta_min = tf.cast(eta_min, tf.float32)
        self.max_step = tf.cast(max_step, tf.float32)
        self.warmup_steps = (
            tf.cast(warmup_steps, tf.float32) if warmup_steps is not None else None
        )

    def __call__(self, step):
        step = tf.cast(step, tf.float32)
        annealing = lambda: (
            self.eta_min
            + 0.5
            * (self.eta_max - self.eta_min)
            * (1 + tf.cos(np.pi * step / self.max_step))
        )
        linear = lambda: self.eta_max * step / tf.cast(self.warmup_steps, tf.float32)
        if self.warmup_steps is None:
            return annealing()
        return tf.cond(
            tf.less(step, self.warmup_steps), true_fn=linear, false_fn=annealing
        )


@register_learning_rate_schedule
class RNMTPlusDecay(tf.keras.optimizers.schedules.LearningRateSchedule):
    """Defines the decay function described in https://arxiv.org/abs/1804.09849."""

    def __init__(
        self, scale, num_replicas, warmup_steps=500, start_step=600000, end_step=1200000
    ):
        """Initializes the decay function.

        Args:
          scale: The scale constant.
          num_replicas: The number of concurrent model replicas.
          warmup_steps: The number of warmup steps.
          start_step: The start step of the exponential decay.
          end_step: The end step of the exponential decay.
        """
        self.scale = tf.cast(scale, tf.float32)
        self.num_replicas = tf.cast(num_replicas, tf.float32)
        self.warmup_steps = tf.cast(warmup_steps, tf.float32)
        self.start_step = tf.cast(start_step, tf.float32)
        self.end_step = tf.cast(end_step, tf.float32)

    def __call__(self, step):
        t = tf.cast(step, tf.float32)
        n = self.num_replicas
        p = self.warmup_steps
        s = self.start_step
        e = self.end_step
        return self.scale * tf.minimum(
            tf.minimum(1 + (t * (n - 1)) / (n * p), n),
            n * tf.pow(2 * n, (s - n * t) / (e - s)),
        )



"""ARK data file to TFRecords converter.

The scripts takes the ARK data file and optionally the indexed target text
to write aligned source and target data.
"""

import argparse

import numpy as np
import tensorflow as tf

from opennmt.inputters.record_inputter import write_sequence_record


def consume_next_vector(ark_file):
    """Consumes the next vector.

    Args:
      ark_file: The ARK data file.

    Returns:
      The next vector as a 2D Numpy array.
    """
    idx = None
    vector = []

    for line in ark_file:
        line = line.strip()
        fields = line.split()

        if not idx:
            idx = fields[0]
            fields.pop(0)
            fields.pop(0)

        end = fields and fields[-1] == "]"

        if end:
            fields.pop()

        if fields:
            vector.append(fields)

        if end:
            break

    return idx, np.asarray(vector, dtype=np.float32)


def consume_next_text(text_file):
    """Consumes the next text line from `text_file`."""
    idx = None
    text = text_file.readline()

    if text:
        tokens = text.strip().split()
        idx = tokens[0]
        tokens.pop(0)
        text = " ".join(tokens)

    return idx, text


def write_text(text, writer):
    """Serializes a line of text."""
    writer.write(text)
    writer.write("\n")


def ark_to_records_aligned(
    ark_filename, text_filename, out_prefix, compression_type=None
):
    """Converts ARK and text datasets to aligned TFRecords and text datasets."""
    record_filename = "%s.records" % out_prefix
    if compression_type == "GZIP":
        record_filename = "%s.gz" % record_filename
    record_writer = tf.io.TFRecordWriter(record_filename, options=compression_type)
    text_writer = open(out_prefix + ".txt", encoding="utf-8", mode="w")

    ark_buffer = {}
    text_buffer = {}
    count = 0

    def _write_example(vector, text):
        write_sequence_record(vector, record_writer)
        write_text(text, text_writer)

    def _search_aligned():
        for idx in ark_buffer:
            if idx in text_buffer:
                vector = ark_buffer[idx]
                text = text_buffer[idx]

                del ark_buffer[idx]
                del text_buffer[idx]

                return vector, text

        return None, None

    with open(ark_filename, encoding="utf-8") as ark_file, open(
        text_filename, encoding="utf-8"
    ) as text_file:
        while True:
            ark_idx, vector = consume_next_vector(ark_file)
            text_idx, text = consume_next_text(text_file)

            if not ark_idx and not text_idx:
                # Both files are empty.
                break

            if ark_idx == text_idx:
                # If the indices match, write the example.
                _write_example(vector, text)
                count += 1
            else:
                # Otherwise store the entries.
                if ark_idx:
                    ark_buffer[ark_idx] = vector
                if text_idx:
                    text_buffer[text_idx] = text

                # Look if we can now find aligned entries.
                vector, text = _search_aligned()

                if vector is not None:
                    _write_example(vector, text)
                    count += 1

    # Search alignments in stored entries.
    while True:
        vector, text = _search_aligned()
        if vector is None:
            break
        _write_example(vector, text)
        count += 1

    record_writer.close()
    text_writer.close()

    print("Saved {} aligned records.".format(count))


def ark_to_records(ark_filename, out_prefix, compression_type=None):
    """Converts ARK dataset to TFRecords."""
    record_writer = tf.io.TFRecordWriter(
        out_prefix + ".records", options=compression_type
    )
    count = 0

    with open(ark_filename, encoding="utf-8") as ark_file:
        while True:
            ark_idx, vector = consume_next_vector(ark_file)
            if not ark_idx:
                break
            write_sequence_record(vector, record_writer)
            count += 1

    record_writer.close()
    print("Saved {} records.".format(count))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--ark", required=True, help="Indexed ARK data file.")
    parser.add_argument(
        "--txt",
        help=(
            "Indexed target text data file "
            "(must set it to align source and target files)."
        ),
    )
    parser.add_argument(
        "--out",
        required=True,
        help="Output files prefix (will be suffixed by .records and .txt).",
    )
    parser.add_argument(
        "--compression_type",
        default=None,
        choices=["GZIP"],
        help="Optional compression type.",
    )
    args = parser.parse_args()

    if args.txt:
        ark_to_records_aligned(
            args.ark, args.txt, args.out, compression_type=args.compression_type
        )
    else:
        ark_to_records(args.ark, args.out, compression_type=args.compression_type)


if __name__ == "__main__":
    main()



"""Dynamic decoding utilities."""

import abc
import collections

import tensorflow as tf
import tensorflow_addons as tfa

from opennmt import constants
from opennmt.utils import misc


class Sampler(abc.ABC):
    """Base class for samplers."""

    @abc.abstractmethod
    def __call__(self, scores, num_samples=1):
        """Samples predictions.

        Args:
          scores: The scores to sample from, a tensor of shape
            ``[batch_size, vocab_size]``.
          num_samples: The number of samples per batch to produce.

        Returns:
          A tuple ``(sample_ids, sample_scores)``.
        """
        raise NotImplementedError()

    @staticmethod
    def from_params(params):
        """Constructs a sampler based on user parameters.

        Args:
          params: A dictionary of user parameters.

        Returns:
          A :class:`opennmt.utils.Sampler` instance.
        """
        sampling_topk = params.get("sampling_topk", 1)
        if sampling_topk == 1:
            return BestSampler()
        else:
            return RandomSampler(
                from_top_k=sampling_topk, temperature=params.get("sampling_temperature")
            )


class RandomSampler(Sampler):
    """Randomly samples from model outputs."""

    def __init__(self, from_top_k=None, temperature=None):
        """Initializes the random sampler.

        Args:
          from_top_k: Sample from the top K predictions instead of the full
            distribution.
          temperature: Divide logits by this value. High temperatures generate more
            random samples.
        """
        if from_top_k is not None and from_top_k <= 0:
            from_top_k = None
        self.from_top_k = from_top_k
        self.temperature = temperature

    def __call__(self, scores, num_samples=1):
        if self.from_top_k is None:
            sample_ids = _sample_from(scores, num_samples, temperature=self.temperature)
        else:
            top_scores, top_ids = tf.nn.top_k(scores, k=self.from_top_k)
            sample_ids = _sample_from(
                top_scores, num_samples, temperature=self.temperature
            )
            sample_ids = _gather_from_word_indices(top_ids, sample_ids)
        sample_scores = _gather_from_word_indices(scores, sample_ids)
        return sample_ids, sample_scores


class BestSampler(Sampler):
    """Sample the best predictions."""

    def __call__(self, scores, num_samples=1):
        sample_scores, sample_ids = tf.nn.top_k(scores, k=num_samples)
        return sample_ids, sample_scores


class DecodingStrategy(abc.ABC):
    """Base class for decoding strategies."""

    @property
    def num_hypotheses(self):
        """The number of hypotheses returned by this strategy."""
        return 1

    @staticmethod
    def from_params(params, tflite_mode=False):
        """Constructs a decoding strategy based on user parameters.

        Args:
          params: A dictionary of user parameters.
          tflite_mode: boolean, should be set to True only if you're exporting with TensorFlow Lite

        Returns:
          A :class:`opennmt.utils.DecodingStrategy` instance.
        """
        beam_size = params.get("beam_width", 1)
        if beam_size > 1:
            return BeamSearch(
                beam_size,
                length_penalty=params.get("length_penalty", 0),
                coverage_penalty=params.get("coverage_penalty", 0),
                tflite_output_size=params.get("tflite_output_size", 250)
                if tflite_mode
                else None,
            )
        else:
            return GreedySearch()

    @abc.abstractmethod
    def initialize(self, start_ids, attention_size=None):
        """Initializes the strategy.

        Args:
          start_ids: The start decoding ids.
          attention_size: If known, the size of the attention vectors (i.e. the
            maximum source length).

        Returns:
          A tuple containing,

          - The (possibly transformed) start decoding ids.
          - The tensor of finished flags.
          - Initial log probabilities per batch.
          - An dictionary of additional tensors used during the decoding.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def step(
        self,
        step,
        sampler,
        log_probs,
        cum_log_probs,
        finished,
        state=None,
        attention=None,
        **kwargs
    ):
        """Updates the strategy state.

        Args:
          step: The current decoding step.
          sampler: The sampler that produces predictions.
          log_probs: The model log probabilities.
          cum_log_probs: The cumulated log probabilities per batch.
          finished: The current finished flags.
          state: The decoder state.
          attention: The attention vector for the current step.
          **kwargs: Additional tensors used by this decoding strategy.

        Returns:
          A tuple containing,

          - The predicted word ids.
          - The new cumulated log probabilities.
          - The updated finished flags.
          - The updated decoder state.
          - A dictionary with additional tensors used by this decoding strategy.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def finalize(self, outputs, end_id, attention=None, **kwargs):
        """Finalize the predictions.

        Args:
          outputs: The array of sampled ids.
          end_id: The end token id.
          attention: The array of attention outputs.
          **kwargs: Additional tensors used by this decoding strategy.

        Returns:
          A tuple containing,

          - The final predictions as a tensor of shape :math:`[B, H, T_t]`.
          - The final attention history of shape :math:`[B, H, T_t, T_s]`.
          - The final sequence lengths of shape :math:`[B, H]`.
        """
        raise NotImplementedError()


class GreedySearch(DecodingStrategy):
    """A basic greedy search strategy."""

    def initialize(self, start_ids, attention_size=None):
        batch_size = tf.shape(start_ids)[0]
        finished = tf.zeros([batch_size], dtype=tf.bool)
        initial_log_probs = tf.zeros([batch_size], dtype=tf.float32)
        return start_ids, finished, initial_log_probs, {}

    def step(
        self,
        step,
        sampler,
        log_probs,
        cum_log_probs,
        finished,
        state=None,
        attention=None,
        **kwargs
    ):
        sample_ids, sample_log_probs = sampler(log_probs)
        sample_ids = tf.reshape(sample_ids, [-1])
        sample_log_probs = tf.reshape(sample_log_probs, [-1])
        cum_log_probs += sample_log_probs
        return sample_ids, cum_log_probs, finished, state, kwargs

    def finalize(self, outputs, end_id, attention=None, **kwargs):
        ids = tf.transpose(outputs.stack())
        ids = tf.expand_dims(ids, 1)
        lengths = _lengths_from_ids(ids, end_id)
        if attention is not None:
            attention = tf.transpose(attention.stack(), perm=[1, 0, 2])
            attention = tf.expand_dims(attention, 1)
        return ids, attention, lengths


class BeamSearch(DecodingStrategy):
    """A beam search strategy."""

    def __init__(
        self, beam_size, length_penalty=0, coverage_penalty=0, tflite_output_size=None
    ):
        """Initializes the decoding strategy.

        Args:
          beam_size: The number of paths to consider per batch.
          length_penalty: Length penalty, see https://arxiv.org/abs/1609.08144.
          coverage_penalty: Coverage penalty, see https://arxiv.org/abs/1609.08144.
          tflite_output_size: None if not TFLite exporting.  Is the output size of TFLite model
        """
        self.beam_size = beam_size
        self.length_penalty = length_penalty
        self.coverage_penalty = coverage_penalty
        self._state_reorder_flags = None
        self.tflite_output_size = tflite_output_size

    @property
    def num_hypotheses(self):
        return self.beam_size

    def _set_state_reorder_flags(self, state_reorder_flags):
        """Sets state reorder flags, a structure matching the decoder state that
        indicates which tensor should be reorded during beam search.
        """
        self._state_reorder_flags = state_reorder_flags

    def initialize(self, start_ids, attention_size=None):
        batch_size = tf.shape(start_ids)[0]
        start_ids = tfa.seq2seq.tile_batch(start_ids, self.beam_size)
        finished = tf.zeros([batch_size * self.beam_size], dtype=tf.bool)
        # Give all probability to first beam for the first iteration.
        initial_log_probs = tf.tile(
            [0.0] + [-float("inf")] * (self.beam_size - 1), [batch_size]
        )
        if self.tflite_output_size is not None:
            parent_ids = tf.TensorArray(
                tf.int32,
                size=self.tflite_output_size,
                dynamic_size=False,
                element_shape=tf.TensorShape(None),
            )
        else:
            parent_ids = tf.TensorArray(tf.int32, size=0, dynamic_size=True)
        extra_vars = {
            "parent_ids": parent_ids,
            "sequence_lengths": tf.zeros([batch_size * self.beam_size], dtype=tf.int32),
        }
        if self.coverage_penalty != 0:
            if attention_size is None:
                raise ValueError(
                    "The attention size should be known to support coverage penalty"
                )
            extra_vars["accumulated_attention"] = tf.zeros(
                [batch_size * self.beam_size, attention_size]
            )
        return start_ids, finished, initial_log_probs, extra_vars

    def _get_scores(
        self, log_probs, sequence_lengths, finished, accumulated_attention=None
    ):
        scores = log_probs
        if self.length_penalty != 0:
            expand_sequence_lengths = tf.expand_dims(sequence_lengths, 1)
            scores /= tf.pow(
                ((5.0 + tf.cast(expand_sequence_lengths + 1, scores.dtype)) / 6.0),
                self.length_penalty,
            )
        if self.coverage_penalty != 0:
            # Mask out of range steps with ones (log(1) == 0).
            accumulated_attention = tf.where(
                tf.equal(accumulated_attention, 0.0),
                x=tf.ones_like(accumulated_attention),
                y=accumulated_attention,
            )
            coverage_penalty = tf.reduce_sum(
                tf.math.log(tf.minimum(accumulated_attention, 1.0)), 1
            )
            # Apply coverage penalty to finished predictions.
            coverage_penalty *= tf.cast(finished, coverage_penalty.dtype)
            scores += self.coverage_penalty * tf.expand_dims(coverage_penalty, 1)
        return scores

    def step(
        self,
        step,
        sampler,
        log_probs,
        cum_log_probs,
        finished,
        state=None,
        attention=None,
        **kwargs
    ):
        parent_ids = kwargs["parent_ids"]
        sequence_lengths = kwargs["sequence_lengths"]

        if self.coverage_penalty != 0:
            if attention is None:
                raise ValueError(
                    "Coverage penalty is enabled but the model did not "
                    "return an attention vector"
                )
            not_finished = tf.math.logical_not(finished)
            attention *= tf.expand_dims(tf.cast(not_finished, attention.dtype), 1)
            accumulated_attention = kwargs["accumulated_attention"] + attention
        else:
            accumulated_attention = None

        # Compute scores from log probabilities.
        vocab_size = log_probs.shape[-1]
        total_probs = log_probs + tf.expand_dims(
            cum_log_probs, 1
        )  # Add current beam probability.
        scores = self._get_scores(
            total_probs,
            sequence_lengths,
            finished,
            accumulated_attention=accumulated_attention,
        )
        scores = tf.reshape(scores, [-1, self.beam_size * vocab_size])
        total_probs = tf.reshape(total_probs, [-1, self.beam_size * vocab_size])

        # Sample predictions.
        sample_ids, sample_scores = sampler(scores, num_samples=self.beam_size)
        cum_log_probs = tf.reshape(
            _gather_from_word_indices(total_probs, sample_ids), [-1]
        )
        sample_ids = tf.reshape(sample_ids, [-1])
        sample_scores = tf.reshape(sample_scores, [-1])

        # Resolve beam origin and word ids.
        word_ids = sample_ids % vocab_size
        beam_ids = sample_ids // vocab_size
        beam_indices = (
            tf.range(tf.shape(word_ids)[0]) // self.beam_size
        ) * self.beam_size + beam_ids

        # Update sequence_length of unfinished sequence.
        sequence_lengths = tf.where(
            finished, x=sequence_lengths, y=sequence_lengths + 1
        )

        # Update state and flags.
        finished = tf.gather(finished, beam_indices)
        sequence_lengths = tf.gather(sequence_lengths, beam_indices)
        parent_ids = parent_ids.write(step, beam_ids)
        extra_vars = {
            "parent_ids": parent_ids,
            "sequence_lengths": sequence_lengths,
        }
        if accumulated_attention is not None:
            extra_vars["accumulated_attention"] = tf.gather(
                accumulated_attention, beam_indices
            )
        if state is not None:
            state = _reorder_state(
                state, beam_indices, reorder_flags=self._state_reorder_flags
            )
        return word_ids, cum_log_probs, finished, state, extra_vars

    def finalize(self, outputs, end_id, attention=None, **kwargs):
        parent_ids = kwargs["parent_ids"]
        sequence_lengths = kwargs["sequence_lengths"]
        maximum_lengths = tf.reduce_max(
            tf.reshape(sequence_lengths, [-1, self.beam_size]), axis=-1
        )
        max_time = outputs.size()
        array_shape = [max_time, -1, self.beam_size]
        step_ids = tf.reshape(outputs.stack(), array_shape)
        parent_ids = tf.reshape(parent_ids.stack(), array_shape)
        ids = _gather_tree(step_ids, parent_ids, maximum_lengths, end_id)
        ids = tf.transpose(ids, perm=[1, 2, 0])
        lengths = _lengths_from_ids(ids, end_id)
        if attention is not None:
            attention = _gather_tree_from_array(attention.stack(), parent_ids, lengths)
            attention = tf.transpose(attention, perm=[1, 0, 2])
            attention = tf.reshape(
                attention, [tf.shape(ids)[0], self.beam_size, max_time, -1]
            )
        return ids, attention, lengths


class DecodingResult(
    collections.namedtuple(
        "DecodingResult", ("ids", "lengths", "log_probs", "attention", "state")
    )
):
    """Final decoding result.

    Args:
      ids: The predicted ids of shape :math:`[B, H, T_t]`.
      lengths: The produced sequences length of shape :math:`[B, H]`.
      log_probs: The cumulated log probabilities of shape :math:`[B, H]`.
      attention: The attention history of shape :math:`[B, H, T_t, T_s]`.
      state: The final decoding state.
    """


def dynamic_decode(
    symbols_to_logits_fn,
    start_ids,
    end_id=constants.END_OF_SENTENCE_ID,
    initial_state=None,
    decoding_strategy=None,
    sampler=None,
    maximum_iterations=None,
    minimum_iterations=0,
    attention_history=False,
    attention_size=None,
    tflite_output_size=None,
):
    """Dynamic decoding.

    Args:
      symbols_to_logits_fn: A callable taking ``(symbols, step, state)`` and
        returning ``(logits, state, attention)`` (``attention`` is optional).
      start_ids: Initial input IDs of shape :math:`[B]`.
      end_id: ID of the end of sequence token.
      initial_state: Initial decoder state.
      decoding_strategy: A :class:`opennmt.utils.DecodingStrategy`
        instance that defines the decoding logic. Defaults to a greedy search.
      sampler: A :class:`opennmt.utils.Sampler` instance that samples
        predictions from the model output. Defaults to an argmax sampling.
      maximum_iterations: The maximum number of iterations to decode for.
      minimum_iterations: The minimum number of iterations to decode for.
      attention_history: Gather attention history during the decoding.
      attention_size: If known, the size of the attention vectors (i.e. the
        maximum source length).
      tflite_output_size: If not None will run TFLite safe, is the size of 1D output tensor.

    Returns:
      A :class:`opennmt.utils.DecodingResult` instance.
    """
    if initial_state is None:
        initial_state = {}
    if decoding_strategy is None:
        decoding_strategy = GreedySearch()
    if sampler is None:
        sampler = BestSampler()
    is_tflite_run = tflite_output_size is not None

    def _cond(
        step, finished, state, inputs, outputs, attention, cum_log_probs, extra_vars
    ):
        return tf.reduce_any(tf.logical_not(finished))

    def _body(
        step, finished, state, inputs, outputs, attention, cum_log_probs, extra_vars
    ):
        # Get log probs from the model.
        result = symbols_to_logits_fn(inputs, step, state)
        logits, state = result[0], result[1]
        attn = result[2] if len(result) > 2 else None
        logits = tf.cast(logits, tf.float32)

        # Penalize or force EOS.
        batch_size, vocab_size = misc.shape_list(logits)
        eos_max_prob = tf.one_hot(
            tf.fill([batch_size], end_id),
            vocab_size,
            on_value=logits.dtype.max,
            off_value=logits.dtype.min,
        )
        logits = tf.cond(
            step < minimum_iterations,
            true_fn=lambda: _penalize_token(logits, end_id),
            false_fn=lambda: tf.where(
                tf.broadcast_to(tf.expand_dims(finished, -1), tf.shape(logits)),
                x=eos_max_prob,
                y=logits,
            ),
        )
        log_probs = tf.nn.log_softmax(logits)

        # Run one decoding strategy step.
        (
            output,
            next_cum_log_probs,
            finished,
            state,
            extra_vars,
        ) = decoding_strategy.step(
            step,
            sampler,
            log_probs,
            cum_log_probs,
            finished,
            state=state,
            attention=attn,
            **extra_vars,
        )

        # Update loop vars.
        outputs = outputs.write(step, output)
        if attention_history:
            if attn is None:
                raise ValueError(
                    "attention_history is set but the model did not return attention"
                )
            attention = attention.write(step, tf.cast(attn, tf.float32))
        cum_log_probs = tf.where(finished, x=cum_log_probs, y=next_cum_log_probs)
        finished = tf.logical_or(finished, tf.equal(output, end_id))
        return (
            step + 1,
            finished,
            state,
            output,
            outputs,
            attention,
            cum_log_probs,
            extra_vars,
        )

    start_ids = tf.convert_to_tensor(start_ids)
    ids_dtype = start_ids.dtype
    start_ids = tf.cast(start_ids, tf.int32)
    start_ids, finished, initial_log_probs, extra_vars = decoding_strategy.initialize(
        start_ids, attention_size=attention_size
    )
    step = tf.constant(0, dtype=tf.int32)

    if is_tflite_run:
        output_shape = tf.TensorShape(None)
        outputs = tf.TensorArray(
            tf.int32,
            size=tflite_output_size,
            dynamic_size=False,
            element_shape=output_shape,
        )
        attn_shape = tf.TensorShape(None)
        attention = tf.TensorArray(
            tf.float32,
            size=tflite_output_size,
            dynamic_size=False,
            element_shape=attn_shape,
        )
        maximum_iterations = (
            tflite_output_size
            if maximum_iterations > tflite_output_size
            else maximum_iterations
        )
    else:
        output_shape = tf.TensorShape(None)
        outputs = tf.TensorArray(tf.int32, size=0, dynamic_size=True)
        attn_shape = tf.TensorShape(None)
        attention = tf.TensorArray(tf.float32, size=0, dynamic_size=True)

    _, _, state, _, outputs, attention, log_probs, extra_vars = tf.while_loop(
        _cond,
        _body,
        loop_vars=(
            step,
            finished,
            initial_state,
            start_ids,
            outputs,
            attention,
            initial_log_probs,
            extra_vars,
        ),
        shape_invariants=(
            step.shape,
            finished.shape,
            tf.nest.map_structure(_get_shape_invariants, initial_state),
            start_ids.shape,
            output_shape,
            attn_shape,
            initial_log_probs.shape,
            tf.nest.map_structure(_get_shape_invariants, extra_vars),
        ),
        parallel_iterations=1,
        maximum_iterations=maximum_iterations,
    )
    ids, attention, lengths = decoding_strategy.finalize(
        outputs,
        end_id,
        attention=attention if attention_history else None,
        **extra_vars,
    )
    if attention is not None:
        attention = attention[:, :, :-1]  # Ignore attention for </s>.
    log_probs = tf.reshape(log_probs, [-1, decoding_strategy.num_hypotheses])
    ids = tf.cast(ids, ids_dtype)
    return DecodingResult(
        ids=ids, lengths=lengths, log_probs=log_probs, attention=attention, state=state
    )


def _reorder_state(state, indices, reorder_flags=None):
    """Gather batch indices from the state tensors."""

    def _reorder_one(tensor, reorder=True):
        if not reorder or isinstance(tensor, tf.TensorArray) or tensor.shape.ndims == 0:
            return tensor
        return tf.gather(tensor, indices)

    args = [state]
    if reorder_flags is not None:
        tf.nest.assert_same_structure(state, reorder_flags)
        args.append(reorder_flags)
    return tf.nest.map_structure(_reorder_one, *args)


def _get_shape_invariants(tensor):
    """Returns the shape of the tensor but sets middle dims to None."""
    if isinstance(tensor, tf.TensorArray):
        shape = None
    else:
        shape = tensor.shape.as_list()
        for i in range(1, len(shape) - 1):
            shape[i] = None
    return tf.TensorShape(shape)


def _penalize_token(log_probs, token_id, penalty=-1e7):
    """Penalize token probabilities."""
    depth = log_probs.shape[-1]
    penalty = tf.one_hot([token_id], depth, on_value=tf.cast(penalty, log_probs.dtype))
    return log_probs + penalty


def _sample_from(logits, num_samples, temperature=None):
    """Sample N values from the unscaled probability distribution."""
    if temperature is not None:
        logits /= tf.cast(temperature, logits.dtype)
    return tf.random.categorical(logits, num_samples, dtype=tf.int32)


def _gather_from_word_indices(tensor, indices):
    """Index the depth dim of a 2D tensor."""
    return tf.gather(tensor, indices, axis=-1, batch_dims=1)


def _lengths_from_ids(ids, end_id):
    """Compute sequence lengths from word ids."""
    lengths = tf.not_equal(ids, end_id)
    lengths = tf.cast(lengths, tf.int32)
    lengths = tf.reduce_sum(lengths, axis=-1)
    return lengths


# The gather_tree functions are imported from TensorFlow Addons:
# https://github.com/tensorflow/addons/blob/master/tensorflow_addons/seq2seq/beam_search_decoder.py
#
# We do not use the Addons version because the public gather_tree function is
# wrapped by a tf.function. This should not be an issue, but the function is
# unexpectedly garbage collected in our test suite.


def _gather_tree(step_ids, parent_ids, max_sequence_lengths, end_token):
    """Calculates the full beams from the per-step ids and parent beam ids.

    For a given beam, past the time step containing the first decoded
    ``end_token`` all values are filled in with ``end_token``.

    Args:
      step_ids: The predicted token IDs.
        A ``int32`` Tensor of shape ``[max_time, batch_size, beam_width]``.
      parent_ids: The parent beam indices.
        A ``int32`` Tensor of shape ``[max_time, batch_size, beam_width]``.
      max_sequence_lengths: The maximum sequence length of each batch.
        A ``int32`` Tensor of shape ``[batch_size]``.
      end_token: The end token ID.

    Returns:
      The reordered token IDs based on ``parent_ids``.

    Raises:
      InvalidArgumentError: if ``parent_ids`` contains an invalid index.
    """
    input_shape = tf.shape(parent_ids)
    max_time = input_shape[0]
    beam_width = input_shape[2]
    max_sequence_lengths = tf.math.minimum(max_sequence_lengths, max_time)
    mask = tf.expand_dims(
        tf.transpose(tf.sequence_mask(max_sequence_lengths, maxlen=max_time)), -1
    )

    # Mask out of range ids.
    end_tokens = tf.fill(input_shape, end_token)
    step_ids = tf.where(mask, x=step_ids, y=end_tokens)
    parent_ids = tf.where(mask, x=parent_ids, y=tf.zeros_like(parent_ids))
    assert_op = tf.debugging.Assert(
        tf.math.reduce_all(
            tf.math.logical_and(parent_ids >= 0, parent_ids < beam_width)
        ),
        ["All parent ids must be positive and less than beam_width"],
    )

    # Reverse all sequences as we need to gather from the end.
    with tf.control_dependencies([assert_op]):
        rev_step_ids = tf.reverse_sequence(
            step_ids, max_sequence_lengths, seq_axis=0, batch_axis=1
        )
        rev_parent_ids = tf.reverse_sequence(
            parent_ids, max_sequence_lengths, seq_axis=0, batch_axis=1
        )

    # Initialize output ids and parent based on last step.
    output_ids = tf.TensorArray(step_ids.dtype, size=max_time, dynamic_size=False)
    output_ids = output_ids.write(0, rev_step_ids[0])
    parent = rev_parent_ids[0]

    # For each step, gather ids based on beam origin.
    for t in tf.range(1, max_time):
        ids = tf.gather(rev_step_ids[t], parent, batch_dims=1)
        parent = tf.gather(rev_parent_ids[t], parent, batch_dims=1)
        output_ids = output_ids.write(t, ids)

    # Reverse sequences to their original order.
    output_ids = output_ids.stack()
    output_ids = tf.reverse_sequence(
        output_ids, max_sequence_lengths, seq_axis=0, batch_axis=1
    )

    # Ensure that there are only end_token after the first end_token.
    in_bound_steps = tf.math.cumsum(tf.cast(output_ids == end_token, tf.int32)) == 0
    output_ids = tf.where(in_bound_steps, x=output_ids, y=end_tokens)
    return output_ids


def _gather_tree_from_array(t, parent_ids, sequence_length):
    """Calculates the full beams for a ``TensorArray``.

    Args:
      t: A stacked ``TensorArray`` of size ``max_time`` that contains Tensors of
        shape ``[batch_size, beam_width, s]`` or ``[batch_size * beam_width, s]``
        where ``s`` is the depth shape.
      parent_ids: The parent ids of shape ``[max_time, batch_size, beam_width]``.
      sequence_length: The sequence length of shape ``[batch_size, beam_width]``.

    Returns:
      A Tensor which is a stacked ``TensorArray`` of the same size and type as
      ``t`` and where beams are sorted in each Tensor according to
      ``parent_ids``.
    """
    max_time = parent_ids.shape[0] or tf.shape(parent_ids)[0]
    batch_size = parent_ids.shape[1] or tf.shape(parent_ids)[1]
    beam_width = parent_ids.shape[2] or tf.shape(parent_ids)[2]

    # Generate beam ids that will be reordered by gather_tree.
    beam_ids = tf.reshape(tf.range(beam_width), [1, 1, -1])
    beam_ids = tf.tile(beam_ids, [max_time, batch_size, 1])

    max_sequence_lengths = tf.cast(tf.reduce_max(sequence_length, axis=1), tf.int32)
    sorted_beam_ids = _gather_tree(
        step_ids=beam_ids,
        parent_ids=parent_ids,
        max_sequence_lengths=max_sequence_lengths,
        end_token=beam_width + 1,
    )

    # For out of range steps, simply copy the same beam.
    in_bound_steps = tf.transpose(
        tf.sequence_mask(sequence_length, maxlen=max_time), perm=[2, 0, 1]
    )
    sorted_beam_ids = tf.where(in_bound_steps, x=sorted_beam_ids, y=beam_ids)

    # Gather from a tensor with collapsed additional dimensions.
    final_shape = tf.shape(t)
    gather_from = tf.reshape(t, [max_time, batch_size, beam_width, -1])
    ordered = tf.gather(gather_from, sorted_beam_ids, axis=2, batch_dims=2)
    ordered = tf.reshape(ordered, final_shape)

    return ordered



"""Define the self-attention encoder."""

import tensorflow as tf

from opennmt.encoders.encoder import Encoder
from opennmt.layers import common, transformer
from opennmt.layers.position import SinusoidalPositionEncoder


class SelfAttentionEncoder(Encoder):
    """Encoder using self-attention as described in
    https://arxiv.org/abs/1706.03762.
    """

    def __init__(
        self,
        num_layers,
        num_units=512,
        num_heads=8,
        ffn_inner_dim=2048,
        dropout=0.1,
        attention_dropout=0.1,
        ffn_dropout=0.1,
        ffn_activation=tf.nn.relu,
        position_encoder_class=SinusoidalPositionEncoder,
        maximum_relative_position=None,
        pre_norm=True,
        **kwargs
    ):
        """Initializes the parameters of the encoder.

        Args:
          num_layers: The number of layers.
          num_units: The number of hidden units.
          num_heads: The number of heads in the multi-head attention.
          ffn_inner_dim: The number of units of the inner linear transformation
            in the feed forward layer.
          dropout: The probability to drop units from the outputs.
          attention_dropout: The probability to drop units from the attention.
          ffn_dropout: The probability to drop units from the activation output in
            the feed forward layer.
          ffn_activation: The activation function to apply between the two linear
            transformations of the feed forward layer.
          position_encoder_class: The :class:`opennmt.layers.PositionEncoder`
            class to use for position encoding (or a callable that returns an
            instance).
          maximum_relative_position: Maximum relative position representation
            (from https://arxiv.org/abs/1803.02155).
          pre_norm: If ``True``, layer normalization is applied before each
            sub-layer. Otherwise it is applied after.
          **kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        self.num_units = num_units
        self.dropout = dropout
        self.position_encoder = None
        if position_encoder_class is not None:
            self.position_encoder = position_encoder_class()
        self.layer_norm = common.LayerNorm() if pre_norm else None
        self.layers = [
            transformer.SelfAttentionEncoderLayer(
                num_units,
                num_heads,
                ffn_inner_dim,
                dropout=dropout,
                attention_dropout=attention_dropout,
                ffn_dropout=ffn_dropout,
                ffn_activation=ffn_activation,
                maximum_relative_position=maximum_relative_position,
                pre_norm=pre_norm,
            )
            for i in range(num_layers)
        ]

    def call(self, inputs, sequence_length=None, training=None):
        inputs *= self.num_units**0.5
        if self.position_encoder is not None:
            inputs = self.position_encoder(inputs)
        inputs = common.dropout(inputs, self.dropout, training=training)
        mask = self.build_mask(inputs, sequence_length=sequence_length)
        for layer in self.layers:
            inputs = layer(inputs, mask=mask, training=training)
        outputs = self.layer_norm(inputs) if self.layer_norm is not None else inputs
        return outputs, None, sequence_length

    def map_v1_weights(self, weights):
        m = []
        m += self.layer_norm.map_v1_weights(weights["LayerNorm"])
        for i, layer in enumerate(self.layers):
            m += layer.map_v1_weights(weights["layer_%d" % i])
        return m



"""Standalone script to detokenize a corpus."""

import argparse

from opennmt import tokenizers


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--delimiter", default=" ", help="Token delimiter used in text serialization."
    )
    parser.add_argument(
        "--tokenizer_config", default=None, help="Tokenization configuration."
    )
    args = parser.parse_args()

    tokenizer = tokenizers.make_tokenizer(args.tokenizer_config)
    tokenizer.detokenize_stream(delimiter=args.delimiter)


if __name__ == "__main__":
    main()



"""Define layers related to the Google's Transformer model."""

import enum

import numpy as np
import tensorflow as tf

from opennmt.layers import common
from opennmt.utils import misc


def _lower_triangle_mask(sequence_length, maximum_length=None, dtype=tf.bool):
    batch_size = tf.shape(sequence_length)[0]
    if maximum_length is None:
        maximum_length = tf.reduce_max(sequence_length)
    mask = tf.ones([batch_size, maximum_length, maximum_length], dtype=dtype)
    mask = tf.linalg.band_part(mask, -1, 0)
    return mask


def future_mask(sequence_length, maximum_length=None, dtype=tf.bool):
    """Builds the dot product mask for future positions.

    Args:
      sequence_length: The sequence length.
      maximum_length: Optional size of the returned time dimension. Otherwise
        it is the maximum of :obj:`sequence_length`.
      dtype: The type of the mask tensor.

    Returns:
      A ``tf.Tensor`` of type :obj:`dtype` and shape
      ``[batch_size, max_length, max_length]``.
    """
    sequence_mask = tf.sequence_mask(
        sequence_length, maxlen=maximum_length, dtype=dtype
    )
    sequence_mask = tf.expand_dims(sequence_mask, axis=1)
    mask = _lower_triangle_mask(
        sequence_length, maximum_length=maximum_length, dtype=dtype
    )
    if dtype is tf.bool:
        return tf.math.logical_and(mask, sequence_mask)
    else:
        return mask * sequence_mask


def split_heads(inputs, num_heads):
    """Splits a tensor in depth.

    Args:
      inputs: A ``tf.Tensor`` of shape :math:`[B, T, D]`.
      num_heads: The number of heads :math:`H`.

    Returns:
      A ``tf.Tensor`` of shape :math:`[B, H, T, D / H]`.
    """
    shape = misc.shape_list(inputs)
    outputs = tf.reshape(inputs, [shape[0], shape[1], num_heads, shape[2] // num_heads])
    outputs = tf.transpose(outputs, perm=[0, 2, 1, 3])
    return outputs


def combine_heads(inputs):
    """Concatenates heads.

    Args:
      inputs: A ``tf.Tensor`` of shape :math:`[B, H, T, D]`.

    Returns:
      A ``tf.Tensor`` of shape :math:`[B, T, D * H]`.
    """
    shape = misc.shape_list(inputs)
    outputs = tf.transpose(inputs, perm=[0, 2, 1, 3])
    outputs = tf.reshape(outputs, [shape[0], shape[2], shape[1] * shape[3]])
    return outputs


def relative_positions(length, maximum_position, with_cache=False):
    """Builds the relative positions.

    Args:
      length: The maximum length of the sequence.
      maximum_position: The maximum relative position to represent.
      with_cache: Set to ``True`` if this function is called from a multi-head
        attention layer with cache states.

    Returns:
      Positive relative positions with shape :math:`[T or 1, T]`.
    """
    if with_cache:
        distance = tf.expand_dims(tf.range(-length + 1, 1, delta=1), 0)
    else:
        arange = tf.range(length)
        distance = tf.expand_dims(arange, 0) - tf.expand_dims(
            arange, 1
        )  # Distance to the diagonal.
    distance = tf.clip_by_value(distance, -maximum_position, maximum_position)
    return distance + maximum_position  # Return positive indices.


def matmul_with_relative_representations(a, b, transpose_b=False):
    """Multiplies :obj:`a` with the relative representations :obj:`b`.

    Args:
      a: Tensor with shape :math:`[B, H, T, _]`.
      b: Tensor with shape :math:`[T, T, _]`.

    Returns:
      Tensor with shape :math:`[B, H, T, T]`.
    """
    batch, head, time, depth = misc.shape_list(a)
    a = tf.transpose(a, perm=[2, 0, 1, 3])
    a = tf.reshape(a, [time, batch * head, depth])
    c = tf.matmul(a, b, transpose_b=transpose_b)
    c = tf.reshape(c, [time, batch, head, c.shape[-1] or tf.shape(c)[-1]])
    c = tf.transpose(c, perm=[1, 2, 0, 3])
    return c


class FeedForwardNetwork(tf.keras.layers.Layer):
    """Implements the Transformer's "Feed Forward" layer.

    .. math::

        ffn(x) = max(0, x*W_1 + b_1)*W_2 + b_2
    """

    def __init__(
        self, inner_dim, output_dim, dropout=0.1, activation=tf.nn.relu, **kwargs
    ):
        """Initializes this layer.

        Args:
          inner_dim: The number of units of the inner linear transformation.
          output_dim: The number of units of the ouput linear transformation.
          dropout: The probability to drop units from the activation output.
          activation: The activation function to apply between the two linear
            transformations.
          kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        self.inner = common.Dense(inner_dim, activation=activation)
        self.outer = common.Dense(output_dim)
        self.dropout = dropout

    def call(self, inputs, training=None):
        """Runs the layer."""
        inner = self.inner(inputs)
        inner = common.dropout(inner, self.dropout, training=training)
        return self.outer(inner)

    def map_v1_weights(self, weights):
        # V1 used conv1d layers that have a leading dimensions.
        weights = tf.nest.map_structure(np.squeeze, weights)
        m = []
        m += self.inner.map_v1_weights(weights["conv1d"])
        m += self.outer.map_v1_weights(weights["conv1d_1"])
        return m


class MultiHeadAttentionReduction(enum.Enum):
    """Enumeration defining how to reduce multi-head attention matrices into a
    single attention matrix.

    Possible values are:

    * ``NONE``: do not reduce and return all attention heads.
    * ``FIRST_HEAD_LAST_LAYER``: take the first attention head of the last layer.
    * ``AVERAGE_LAST_LAYER``: average of all attention heads of the last layer.
    * ``AVERAGE_ALL_LAYERS``: average of all attention heads.
    """

    NONE = 0
    FIRST_HEAD_LAST_LAYER = 1
    AVERAGE_LAST_LAYER = 2
    AVERAGE_ALL_LAYERS = 3

    @staticmethod
    def reduce(attention, policy):
        """Reduces a list of multi-head attention matrices into a single
        attention matrix.

        Args:
          attention: A list of multi-head attention matrices, with shape
            :math:`[B, H, T_t, T_s]`.
          policy: A :class:`opennmt.layers.MultiHeadAttentionReduction` value.

        Returns:
          A single attention matrix with shape :math:`[B, T_t, T_s]`.
        """
        if not isinstance(attention, list):
            raise ValueError(
                "attention should be a list of attention tensors, one per layer"
            )
        if policy == MultiHeadAttentionReduction.FIRST_HEAD_LAST_LAYER:
            attention = attention[-1][:, 0]
        elif policy == MultiHeadAttentionReduction.AVERAGE_LAST_LAYER:
            attention = tf.math.reduce_mean(attention[-1], axis=1)
        elif policy == MultiHeadAttentionReduction.AVERAGE_ALL_LAYERS:
            attention = tf.concat(attention, axis=1)
            attention = tf.math.reduce_mean(attention, axis=1)
        return attention


class MultiHeadAttention(tf.keras.layers.Layer):
    """Computes the multi-head attention as described in
    https://arxiv.org/abs/1706.03762.
    """

    def __init__(
        self,
        num_heads,
        num_units,
        dropout=0.1,
        return_attention=False,
        maximum_relative_position=None,
        **kwargs
    ):
        """Initializes this layer.

        Args:
          num_heads: The number of attention heads.
          num_units: The number of hidden units.
          dropout: The probability to drop units from the inputs.
          return_attention: If ``True``, also return the attention weights.
          maximum_relative_position: Maximum relative position representation
            (from https://arxiv.org/abs/1803.02155).
          kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        if num_units % num_heads != 0:
            raise ValueError(
                "Multi head attention requires that num_units is a"
                " multiple of %s" % num_heads
            )
        self.num_heads = num_heads
        self.num_units_per_head = num_units // num_heads
        self.linear_queries = common.Dense(num_units)
        self.linear_keys = common.Dense(num_units)
        self.linear_values = common.Dense(num_units)
        self.linear_output = common.Dense(num_units)
        self.dropout = dropout
        self.return_attention = return_attention
        self.maximum_relative_position = maximum_relative_position

    def map_v1_weights(self, weights):
        # V1 used conv1d layers that have a leading dimensions.
        weights = tf.nest.map_structure(np.squeeze, weights)

        # V1 used fused linear projections, so the weights should be split accordingly.
        def _partial_weights(key, num_splits, index):
            return tf.nest.map_structure(
                lambda w: np.split(w, num_splits, axis=0 if w.ndim == 1 else 1)[index],
                weights[key],
            )

        m = []
        if "conv1d_2" not in weights:  # Case self-attention.
            m += self.linear_queries.map_v1_weights(_partial_weights("conv1d", 3, 0))
            m += self.linear_keys.map_v1_weights(_partial_weights("conv1d", 3, 1))
            m += self.linear_values.map_v1_weights(_partial_weights("conv1d", 3, 2))
            m += self.linear_output.map_v1_weights(weights["conv1d_1"])
        else:
            m += self.linear_queries.map_v1_weights(weights["conv1d"])
            m += self.linear_keys.map_v1_weights(_partial_weights("conv1d_1", 2, 0))
            m += self.linear_values.map_v1_weights(_partial_weights("conv1d_1", 2, 1))
            m += self.linear_output.map_v1_weights(weights["conv1d_2"])
        return m

    def build(self, input_shape):
        if self.maximum_relative_position is not None:
            relative_window_size = self.maximum_relative_position * 2 + 1
            relative_repr_shape = [relative_window_size, self.num_units_per_head]
            self.relative_position_keys = self.add_weight(
                name="relative_position_keys", shape=relative_repr_shape
            )
            self.relative_position_values = self.add_weight(
                name="relative_position_values", shape=relative_repr_shape
            )
        super().build(input_shape)

    def call(self, inputs, memory=None, mask=None, cache=None, training=None):
        """Runs the layer.

        Args:
          inputs: The sequence of queries. A tensor of shape :math:`[B, T_1, ...]`.
          memory: The sequence to attend. A tensor of shape :math:`[B, T_2, ...]`.
            If ``None``, computes self-attention.
          mask: The dot product mask. A boolean tensor of shape :math:`[B, T_2]` or
            :math:`[B, T_1, T_2]`.
          cache: An optional tuple containing projected keys and values from the
            previous step. Tensors of shape :math:`[B, H, T_2, D / H]`.
          training: Run in training mode.

        Returns:
          A tuple with the attention context, the updated cache and the attention
          weights (if :obj:`return_attention` is ``True``).
        """

        def _compute_kv(x):
            keys = self.linear_keys(x)
            keys = split_heads(keys, self.num_heads)
            values = self.linear_values(x)
            values = split_heads(values, self.num_heads)
            return keys, values

        # Compute queries.
        queries = self.linear_queries(inputs)
        queries = split_heads(queries, self.num_heads)
        queries *= self.num_units_per_head**-0.5

        # Compute keys and values.
        if memory is None:
            keys, values = _compute_kv(inputs)
            if cache:
                keys = tf.concat([cache[0], keys], axis=2)
                values = tf.concat([cache[1], values], axis=2)
        else:
            if cache:
                keys, values = tf.cond(
                    tf.equal(tf.shape(cache[0])[2], 0),
                    true_fn=lambda: _compute_kv(memory),
                    false_fn=lambda: cache,
                )
            else:
                keys, values = _compute_kv(memory)

        if self.maximum_relative_position is not None:
            if memory is not None:
                raise ValueError(
                    "Relative position representations only supports self-attention"
                )
            keys_length = tf.shape(keys)[2]
            relative_pos = relative_positions(
                keys_length, self.maximum_relative_position, with_cache=bool(cache)
            )

            # Uses gather_nd instead of nn.embedding_lookup for TFLite exporting (TF Issue #42410)
            relative_pos_expanded = tf.expand_dims(relative_pos, axis=-1)
            relative_repr_keys = tf.gather_nd(
                self.relative_position_keys, relative_pos_expanded
            )
            relative_repr_values = tf.gather_nd(
                self.relative_position_values, relative_pos_expanded
            )
        else:
            relative_repr_keys = None
            relative_repr_values = None

        cache = (keys, values)

        # Dot product attention.
        dot = tf.matmul(queries, keys, transpose_b=True)
        if relative_repr_keys is not None:
            dot += matmul_with_relative_representations(
                queries, relative_repr_keys, transpose_b=True
            )
        if mask is not None:
            mask = tf.cast(mask, tf.float32)
            if mask.shape.rank == 2:
                mask = tf.expand_dims(mask, 1)  # Broadcast on time dimension.
            mask = tf.expand_dims(mask, 1)  # Broadcast on head dimension.
            dot = tf.cast(
                tf.cast(dot, tf.float32) * mask + ((1.0 - mask) * tf.float32.min),
                dot.dtype,
            )
        attn = tf.cast(tf.nn.softmax(tf.cast(dot, tf.float32)), dot.dtype)
        drop_attn = common.dropout(attn, self.dropout, training=training)
        heads = tf.matmul(drop_attn, values)
        if relative_repr_values is not None:
            heads += matmul_with_relative_representations(
                drop_attn, relative_repr_values
            )

        # Concatenate all heads output.
        combined = combine_heads(heads)
        outputs = self.linear_output(combined)
        if self.return_attention:
            return outputs, cache, attn
        return outputs, cache


class TransformerLayerWrapper(common.LayerWrapper):
    """Layer wrapper that applies a standard Transformer preprocessing and
    postprocessing.

    With ``pre_norm=True``:

    .. code-block:: text

        y = layer_norm(x)
        y = dropout(layer(y)) + x

    With ``pre_norm=False``:

    .. code-block:: text

        y = dropout(layer(x)) + x
        y = layer_norm(y)
    """

    def __init__(self, layer, output_dropout, pre_norm=True, **kwargs):
        """Initializes the wrapper.

        Args:
          layer: The Transformer layer to wrap.
          output_dropout: The dropout to apply on the layer output.
          pre_norm: If ``True``, layer normalization is applied before calling
            the layer. Otherwise it is applied after.
          **kwargs: Additional layer arguments.
        """
        super().__init__(
            layer,
            normalize_input=pre_norm,
            normalize_output=not pre_norm,
            output_dropout=output_dropout,
            residual_connection=True,
            **kwargs,
        )

    def map_v1_weights(self, weights):
        m = []
        m += self.input_layer_norm.map_v1_weights(weights["LayerNorm"])
        m += self.layer.map_v1_weights(weights)
        return m


class SelfAttentionEncoderLayer(tf.keras.layers.Layer):
    """Implements one self-attention encoding layer."""

    def __init__(
        self,
        num_units,
        num_heads,
        ffn_inner_dim,
        dropout=0.1,
        attention_dropout=0.1,
        ffn_dropout=0.1,
        ffn_activation=tf.nn.relu,
        maximum_relative_position=None,
        pre_norm=True,
        **kwargs
    ):
        """Initializes the layer.

        Args:
          num_units: The number of hidden units.
          num_heads: The number of heads in the multi-head attention.
          ffn_inner_dim: The number of units of the inner linear transformation
            in the feed forward layer.
          dropout: The probability to drop units from the outputs.
          attention_dropout: The probability to drop units from the attention.
          ffn_dropout: The probability to drop units from the activation output in
            the feed forward layer.
          ffn_activation: The activation function to apply between the two linear
            transformations of the feed forward layer.
          maximum_relative_position: Maximum relative position representation
            (from https://arxiv.org/abs/1803.02155).
          pre_norm: If ``True``, layer normalization is applied before each
            sub-layer. Otherwise it is applied after.
          **kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        self.self_attention = MultiHeadAttention(
            num_heads,
            num_units,
            dropout=attention_dropout,
            maximum_relative_position=maximum_relative_position,
        )
        self.self_attention = TransformerLayerWrapper(
            self.self_attention, dropout, pre_norm=pre_norm
        )
        self.ffn = FeedForwardNetwork(
            ffn_inner_dim, num_units, dropout=ffn_dropout, activation=ffn_activation
        )
        self.ffn = TransformerLayerWrapper(self.ffn, dropout, pre_norm=pre_norm)

    def call(self, x, mask=None, training=None):
        """Runs the encoder layer."""
        y, _ = self.self_attention(x, mask=mask, training=training)
        y = self.ffn(y, training=training)
        return y

    def map_v1_weights(self, weights):
        m = []
        m += self.self_attention.map_v1_weights(weights["multi_head"])
        m += self.ffn.map_v1_weights(weights["ffn"])
        return m


class SelfAttentionDecoderLayer(tf.keras.layers.Layer):
    """Implements one self-attention decoding layer."""

    def __init__(
        self,
        num_units,
        num_heads,
        ffn_inner_dim,
        num_sources=1,
        dropout=0.1,
        attention_dropout=0.1,
        ffn_dropout=0.1,
        ffn_activation=tf.nn.relu,
        maximum_relative_position=None,
        pre_norm=True,
        **kwargs
    ):
        """Initializes the layer.

        Args:
          num_units: The number of hidden units.
          num_heads: The number of heads in the multi-head attention.
          ffn_inner_dim: The number of units of the inner linear transformation
            in the feed forward layer.
          num_sources: The number of source contexts.
          dropout: The probability to drop units from the outputs.
          attention_dropout: The probability to drop units from the attention.
          ffn_dropout: The probability to drop units from the activation output in
            the feed forward layer.
          ffn_activation: The activation function to apply between the two linear
            transformations of the feed forward layer.
          maximum_relative_position: Maximum relative position representation
            (from https://arxiv.org/abs/1803.02155).
          pre_norm: If ``True``, layer normalization is applied before each
            sub-layer. Otherwise it is applied after.
          **kwargs: Additional layer arguments.
        """
        super().__init__(**kwargs)
        self.self_attention = MultiHeadAttention(
            num_heads,
            num_units,
            dropout=attention_dropout,
            maximum_relative_position=maximum_relative_position,
        )
        self.self_attention = TransformerLayerWrapper(
            self.self_attention, dropout, pre_norm=pre_norm
        )
        self.attention = []
        for _ in range(num_sources):
            attention = MultiHeadAttention(
                num_heads, num_units, dropout=attention_dropout, return_attention=True
            )
            attention = TransformerLayerWrapper(attention, dropout, pre_norm=pre_norm)
            self.attention.append(attention)
        self.ffn = FeedForwardNetwork(
            ffn_inner_dim, num_units, dropout=ffn_dropout, activation=ffn_activation
        )
        self.ffn = TransformerLayerWrapper(self.ffn, dropout, pre_norm=pre_norm)

    def map_v1_weights(self, weights):
        m = []
        m += self.self_attention.map_v1_weights(weights["masked_multi_head"])
        m += self.attention[0].map_v1_weights(weights["multi_head"])
        m += self.ffn.map_v1_weights(weights["ffn"])
        return m

    def call(
        self,
        inputs,
        mask=None,
        memory=None,
        memory_mask=None,
        cache=None,
        training=None,
    ):
        """Runs the decoder layer."""
        if cache is None:
            cache = {}

        outputs, self_kv = self.self_attention(
            inputs, mask=mask, cache=cache.get("self_kv"), training=training
        )

        attention = []
        memory_kv = []
        if memory is not None:
            memory_cache = cache.get("memory_kv")
            if memory_cache is None:
                memory_cache = [None] * len(self.attention)
            for layer, mem, mem_mask, mem_cache in zip(
                self.attention, memory, memory_mask, memory_cache
            ):
                outputs, memory_kv_i, attention_i = layer(
                    outputs,
                    memory=mem,
                    mask=mem_mask,
                    cache=mem_cache,
                    training=training,
                )
                attention.append(attention_i)
                memory_kv.append(memory_kv_i)

        outputs = self.ffn(outputs, training=training)
        cache = dict(self_kv=self_kv, memory_kv=memory_kv)
        return outputs, cache, attention



"""This module exposes classes and functions that help creating and processing data."""

from opennmt.data.dataset import (
    batch_dataset,
    batch_sequence_dataset,
    filter_examples_by_length,
    filter_irregular_batches,
    get_dataset_size,
    inference_pipeline,
    make_cardinality_multiple_of,
    random_shard,
    shuffle_dataset,
    training_pipeline,
)
from opennmt.data.noise import (
    Noise,
    WordDropout,
    WordNoiser,
    WordOmission,
    WordPermutation,
    WordReplacement,
)
from opennmt.data.text import (
    alignment_matrix_from_pharaoh,
    tokens_to_chars,
    tokens_to_words,
)
from opennmt.data.vocab import Vocab, create_lookup_tables



"""Define position encoder classes."""

import abc
import math

import tensorflow as tf

from opennmt.layers.reducer import SumReducer


class PositionEncoder(tf.keras.layers.Layer):
    """Base class for position encoders."""

    def __init__(self, reducer=None, **kwargs):
        """Initializes the position encoder.

        Args:
          reducer: A :class:`opennmt.layers.Reducer` to merge inputs and position
            encodings. Defaults to :class:`opennmt.layers.SumReducer`.
          **kwargs: Additional layer keyword arguments.
        """
        super().__init__(**kwargs)
        if reducer is None:
            reducer = SumReducer(dtype=kwargs.get("dtype"))
        self.reducer = reducer

    def call(self, inputs, position=None):
        """Add position encodings to :obj:`inputs`.

        Args:
          inputs: The inputs to encode.
          position: The single position to encode, to use when this layer is called
            step by step.

        Returns:
          A ``tf.Tensor`` whose shape depends on the configured ``reducer``.
        """
        batch_size = tf.shape(inputs)[0]
        timesteps = tf.shape(inputs)[1]
        input_dim = inputs.shape[-1]
        positions = tf.range(timesteps) + 1 if position is None else [position]
        position_encoding = self._encode([positions], input_dim)
        position_encoding = tf.tile(position_encoding, [batch_size, 1, 1])
        return self.reducer([inputs, position_encoding])

    @abc.abstractmethod
    def _encode(self, positions, depth):
        """Creates position encodings.

        Args:
          positions: The positions to encode of shape :math:`[B, ...]`.
          depth: The encoding depth :math:`D`.

        Returns:
          A ``tf.Tensor`` of shape :math:`[B, ..., D]`.
        """
        raise NotImplementedError()


class PositionEmbedder(PositionEncoder):
    """Encodes position with a lookup table."""

    def __init__(self, maximum_position=128, reducer=None, **kwargs):
        """Initializes the position encoder.

        Args:
          maximum_position: The maximum position to embed. Positions greater
            than this value will be set to :obj:`maximum_position`.
          reducer: A :class:`opennmt.layers.Reducer` to merge inputs and position
            encodings. Defaults to :class:`opennmt.layers.SumReducer`.
          **kwargs: Additional layer keyword arguments.
        """
        super().__init__(reducer=reducer, **kwargs)
        self.maximum_position = maximum_position
        self.embedding = None

    def build(self, input_shape):
        shape = [self.maximum_position + 1, input_shape[-1]]
        self.embedding = self.add_weight("position_embedding", shape)
        super().build(input_shape)

    def _encode(self, positions, depth):
        positions = tf.minimum(positions, self.maximum_position)

        if getattr(self, "_tflite_mode", False):
            # Workaround for TensorFlow issue #42410.
            positions = tf.expand_dims(positions, axis=-1)
            return tf.gather_nd(self.embedding, positions)

        return tf.nn.embedding_lookup(self.embedding, positions)


class SinusoidalPositionEncoder(PositionEncoder):
    """Encodes positions with sine waves as described in
    https://arxiv.org/abs/1706.03762.
    """

    def _encode(self, positions, depth):
        if depth % 2 != 0:
            raise ValueError(
                "SinusoidalPositionEncoder expects the depth to be divisble "
                "by 2 but got %d" % depth
            )

        batch_size = tf.shape(positions)[0]
        positions = tf.cast(positions, tf.float32)

        log_timescale_increment = math.log(10000) / (depth / 2 - 1)
        inv_timescales = tf.exp(
            tf.range(depth / 2, dtype=tf.float32) * -log_timescale_increment
        )
        inv_timescales = tf.reshape(
            tf.tile(inv_timescales, [batch_size]), [batch_size, depth // 2]
        )
        scaled_time = tf.expand_dims(positions, -1) * tf.expand_dims(inv_timescales, 1)
        encoding = tf.concat([tf.sin(scaled_time), tf.cos(scaled_time)], axis=2)
        return tf.cast(encoding, self.dtype)



"""Inference related classes and functions."""

import sys
import time

import tensorflow as tf

from opennmt.utils import misc


def predict_dataset(
    model, dataset, print_params=None, predictions_file=None, log_time=False
):
    """Outputs the model predictions for the dataset.

    To run inference on strings directly, see
    :meth:`opennmt.models.Model.serve_function`.

    Args:
      model: A :class:`opennmt.models.Model` instance.
      dataset: A ``tf.data.Dataset`` instance outputting features.
      print_params: A dictionary of parameters passed to
        :meth:`opennmt.models.Model.print_prediction`.
      predictions_file: If set, predictions are saved in this file, otherwise they
        are printed on the standard output.
      log_time: If ``True``, several time metrics will be printed in the logs at
        the end of the inference loop.
    """
    if predictions_file:
        stream = open(predictions_file, encoding="utf-8", mode="w")
    else:
        stream = sys.stdout

    infer_fn = tf.function(model.infer, input_signature=(dataset.element_spec,))
    if not tf.config.functions_run_eagerly():
        tf.get_logger().info("Tracing and optimizing the inference graph...")
        infer_fn.get_concrete_function()  # Trace the function now.

    # Inference might return out-of-order predictions. The OrderRestorer utility is
    # used to write predictions in their original order.
    write_fn = lambda prediction: (
        model.print_prediction(prediction, params=print_params, stream=stream)
    )
    index_fn = lambda prediction: prediction.get("index")
    ordered_writer = misc.OrderRestorer(index_fn, write_fn)

    total_time = 0
    total_tokens = 0
    total_examples = 0
    start_time = time.time()

    # When the inference dataset is bucketized, it can happen that no output is
    # written in a long time. To avoid confusion and give the impression that
    # the process is stuck, we ensure that something is logged regularly.
    max_time_without_output = 10
    last_output_time = start_time

    for features in dataset:
        predictions = infer_fn(features)
        predictions = tf.nest.map_structure(lambda t: t.numpy(), predictions)
        batch_time = time.time()

        for prediction in misc.extract_batches(predictions):
            written = ordered_writer.push(prediction)
            if written:
                last_output_time = batch_time
            else:
                time_without_output = batch_time - last_output_time
                if time_without_output >= max_time_without_output:
                    tf.get_logger().info(
                        "%d predictions are buffered, but waiting for the prediction of "
                        "line %d to advance the output...",
                        ordered_writer.buffer_size,
                        ordered_writer.next_index + 1,
                    )
                    last_output_time = batch_time

        if log_time:
            batch_size = next(iter(predictions.values())).shape[0]
            total_examples += batch_size
            length = predictions.get("length")
            if length is not None:
                if len(length.shape) == 2:
                    length = length[:, 0]
                total_tokens += sum(length)

    if log_time:
        end_time = time.time()
        total_time = end_time - start_time
        tf.get_logger().info("Total prediction time (s): %f", total_time)
        tf.get_logger().info(
            "Average prediction time (s): %f", total_time / total_examples
        )
        if total_tokens > 0:
            tf.get_logger().info("Tokens per second: %f", total_tokens / total_time)
    if predictions_file:
        stream.close()


def score_dataset(model, dataset, print_params=None, output_file=None):
    """Outputs the model scores for the dataset.

    Args:
      model: A :class:`opennmt.models.Model` instance.
      dataset: A ``tf.data.Dataset`` instance outputting parallel features and
        labels.
      print_params: A dictionary of parameters passed to
        :meth:`opennmt.models.Model.print_score`.
      output_file: If set, outputs are saved in this file, otherwise they are
        printed on the standard output.
    """
    if output_file:
        stream = open(output_file, encoding="utf-8", mode="w")
    else:
        stream = sys.stdout

    write_fn = lambda batch: (
        model.print_score(batch, params=print_params, stream=stream)
    )
    index_fn = lambda batch: batch.get("index")
    ordered_writer = misc.OrderRestorer(index_fn, write_fn)

    score_fn = tf.function(model.score, input_signature=dataset.element_spec)
    for features, labels in dataset:
        results = score_fn(features, labels)
        results = tf.nest.map_structure(lambda t: t.numpy(), results)
        for batch in misc.extract_batches(results):
            ordered_writer.push(batch)

    if output_file:
        stream.close()



"""OpenNMT module."""

from opennmt.version import __version__, _check_tf_version

_check_tf_version()

from opennmt.config import convert_to_v2_config, load_config, load_model, merge_config
from opennmt.constants import (
    END_OF_SENTENCE_ID,
    END_OF_SENTENCE_TOKEN,
    PADDING_ID,
    PADDING_TOKEN,
    START_OF_SENTENCE_ID,
    START_OF_SENTENCE_TOKEN,
    UNKNOWN_TOKEN,
)
from opennmt.runner import Runner



"""Catalog of predefined models."""

import tensorflow as tf
import tensorflow_addons as tfa

from opennmt import config as config_util
from opennmt import decoders, encoders, inputters, layers
from opennmt.models import (
    language_model,
    model,
    sequence_tagger,
    sequence_to_sequence,
    transformer,
)
from opennmt.utils import misc

_CATALOG_MODELS_REGISTRY = misc.ClassRegistry(base_class=model.Model)

register_model_in_catalog = _CATALOG_MODELS_REGISTRY.register


def list_model_names_from_catalog():
    """Lists the models name registered in the catalog."""
    return _CATALOG_MODELS_REGISTRY.class_names


def get_model_from_catalog(name, as_builder=False):
    """Gets a model from the catalog.

    Args:
      name: The model name in the catalog.
      as_builder: If ``True``, return a callable building the model on call.

    Returns:
      A :class:`opennmt.models.Model` instance or a callable returning such
      instance.

    Raises:
      ValueError: if the model :obj:`name` does not exist in the catalog.
    """
    model_class = _CATALOG_MODELS_REGISTRY.get(name)
    if model_class is None:
        raise ValueError("The model '%s' does not exist in the model catalog" % name)
    if as_builder:
        return model_class
    return model_class()


@register_model_in_catalog
class ListenAttendSpell(sequence_to_sequence.SequenceToSequence):
    """Defines a model similar to the "Listen, Attend and Spell" model described
    in https://arxiv.org/abs/1508.01211.
    """

    def __init__(self):
        super().__init__(
            source_inputter=inputters.SequenceRecordInputter(input_depth=40),
            target_inputter=inputters.WordEmbedder(embedding_size=50),
            encoder=encoders.PyramidalRNNEncoder(
                num_layers=3,
                num_units=512,
                reduction_factor=2,
                cell_class=tf.keras.layers.LSTMCell,
                dropout=0.3,
            ),
            decoder=decoders.AttentionalRNNDecoder(
                num_layers=3,
                num_units=512,
                attention_mechanism_class=tfa.seq2seq.LuongMonotonicAttention,
                cell_class=tf.keras.layers.LSTMCell,
                dropout=0.3,
                residual_connections=False,
                first_layer_attention=True,
            ),
        )

    def auto_config(self, num_replicas=1):
        config = super().auto_config(num_replicas=num_replicas)
        return config_util.merge_config(
            config,
            {
                "params": {
                    "optimizer": "SGD",
                    "learning_rate": 0.2,
                    "scheduled_sampling_type": "constant",
                    "scheduled_sampling_read_probability": 0.9,
                },
                "train": {
                    "batch_size": 32,
                    "length_bucket_width": 15,
                    "maximum_features_length": 2450,
                    "maximum_labels_length": 330,
                },
            },
        )


class _RNNBase(sequence_to_sequence.SequenceToSequence):
    """Base class for RNN based NMT models."""

    def auto_config(self, num_replicas=1):
        config = super().auto_config(num_replicas=num_replicas)
        return config_util.merge_config(
            config,
            {
                "params": {
                    "optimizer": "Adam",
                    "learning_rate": 0.0002,
                },
                "train": {
                    "batch_size": 64,
                    "maximum_features_length": 80,
                    "maximum_labels_length": 80,
                },
            },
        )


@register_model_in_catalog
class LuongAttention(_RNNBase):
    """Defines a LSTM encoder-decoder model as described in https://arxiv.org/abs/1508.04025."""

    def __init__(self):
        super().__init__(
            source_inputter=inputters.WordEmbedder(embedding_size=512),
            target_inputter=inputters.WordEmbedder(embedding_size=512),
            encoder=encoders.RNNEncoder(
                num_layers=4,
                num_units=1000,
                dropout=0.2,
                residual_connections=False,
                cell_class=tf.keras.layers.LSTMCell,
            ),
            decoder=decoders.AttentionalRNNDecoder(
                num_layers=4,
                num_units=1000,
                bridge_class=layers.CopyBridge,
                attention_mechanism_class=tfa.seq2seq.LuongAttention,
                cell_class=tf.keras.layers.LSTMCell,
                dropout=0.2,
                residual_connections=False,
            ),
        )


@register_model_in_catalog
class NMTBigV1(_RNNBase):
    """Defines a bidirectional LSTM encoder-decoder model.

    Note:
      For compatibility with OpenNMT-tf v1.
    """

    def __init__(self):
        super().__init__(
            source_inputter=inputters.WordEmbedder(embedding_size=512),
            target_inputter=inputters.WordEmbedder(embedding_size=512),
            encoder=encoders.RNNEncoder(
                num_layers=4,
                num_units=512,
                bidirectional=True,
                residual_connections=False,
                dropout=0.3,
                reducer=layers.ConcatReducer(),
                cell_class=tf.keras.layers.LSTMCell,
            ),
            decoder=decoders.AttentionalRNNDecoder(
                num_layers=4,
                num_units=1024,
                bridge_class=layers.CopyBridge,
                attention_mechanism_class=tfa.seq2seq.LuongAttention,
                attention_layer_activation=None,
                cell_class=tf.keras.layers.LSTMCell,
                dropout=0.3,
                residual_connections=False,
            ),
        )


@register_model_in_catalog
class NMTMediumV1(_RNNBase):
    """Defines a medium-sized bidirectional LSTM encoder-decoder model.

    Note:
      For compatibility with OpenNMT-tf v1.
    """

    def __init__(self):
        super().__init__(
            source_inputter=inputters.WordEmbedder(embedding_size=512),
            target_inputter=inputters.WordEmbedder(embedding_size=512),
            encoder=encoders.RNNEncoder(
                num_layers=4,
                num_units=256,
                bidirectional=True,
                residual_connections=False,
                dropout=0.3,
                reducer=layers.ConcatReducer(),
                cell_class=tf.keras.layers.LSTMCell,
            ),
            decoder=decoders.AttentionalRNNDecoder(
                num_layers=4,
                num_units=512,
                bridge_class=layers.CopyBridge,
                attention_mechanism_class=tfa.seq2seq.LuongAttention,
                attention_layer_activation=None,
                cell_class=tf.keras.layers.LSTMCell,
                dropout=0.3,
                residual_connections=False,
            ),
        )


@register_model_in_catalog
class NMTSmallV1(_RNNBase):
    """Defines a small unidirectional LSTM encoder-decoder model.

    Note:
      For compatibility with OpenNMT-tf v1.
    """

    def __init__(self):
        super().__init__(
            source_inputter=inputters.WordEmbedder(embedding_size=512),
            target_inputter=inputters.WordEmbedder(embedding_size=512),
            encoder=encoders.RNNEncoder(
                num_layers=2,
                num_units=512,
                residual_connections=False,
                dropout=0.3,
                cell_class=tf.keras.layers.LSTMCell,
            ),
            decoder=decoders.AttentionalRNNDecoder(
                num_layers=2,
                num_units=512,
                bridge_class=layers.CopyBridge,
                attention_mechanism_class=tfa.seq2seq.LuongAttention,
                attention_layer_activation=None,
                cell_class=tf.keras.layers.LSTMCell,
                dropout=0.3,
                residual_connections=False,
            ),
        )


@register_model_in_catalog
class LstmCnnCrfTagger(sequence_tagger.SequenceTagger):
    """Defines a bidirectional LSTM-CNNs-CRF as described in https://arxiv.org/abs/1603.01354."""

    def __init__(self):
        super().__init__(
            inputter=inputters.MixedInputter(
                [
                    inputters.WordEmbedder(embedding_size=100),
                    inputters.CharConvEmbedder(
                        embedding_size=30,
                        num_outputs=30,
                        kernel_size=3,
                        stride=1,
                        dropout=0.5,
                    ),
                ],
                dropout=0.5,
            ),
            encoder=encoders.RNNEncoder(
                num_layers=1,
                num_units=400,
                bidirectional=True,
                dropout=0.5,
                residual_connections=False,
                cell_class=tf.keras.layers.LSTMCell,
            ),
            crf_decoding=True,
        )

    def auto_config(self, num_replicas=1):
        config = super().auto_config(num_replicas=num_replicas)
        return config_util.merge_config(
            config,
            {
                "params": {
                    "optimizer": "Adam",
                    "learning_rate": 0.001,
                },
                "train": {
                    "batch_size": 32,
                },
            },
        )


@register_model_in_catalog(alias="Transformer")
class TransformerBase(transformer.Transformer):
    """Defines a base Transformer model as described in https://arxiv.org/abs/1706.03762."""


@register_model_in_catalog
class TransformerBaseSharedEmbeddings(transformer.Transformer):
    """Defines a base Transformer model with shared embeddings as described in
    https://arxiv.org/abs/1706.03762.
    """

    def __init__(self):
        super().__init__(
            share_embeddings=sequence_to_sequence.EmbeddingsSharingLevel.ALL,
        )


@register_model_in_catalog(alias="TransformerRelative")
class TransformerBaseRelative(transformer.Transformer):
    """Defines a base Transformer model using relative position representations as
    described in https://arxiv.org/abs/1803.02155.
    """

    def __init__(self):
        super().__init__(position_encoder_class=None, maximum_relative_position=20)


# Backward compatibility with model descriptions that directly accessed the catalog module.
Transformer = TransformerBase
TransformerRelative = TransformerBaseRelative


@register_model_in_catalog
class TransformerBig(transformer.Transformer):
    """Defines a big Transformer model as described in https://arxiv.org/abs/1706.03762."""

    def __init__(self):
        super().__init__(num_units=1024, num_heads=16, ffn_inner_dim=4096)


@register_model_in_catalog
class TransformerBigSharedEmbeddings(transformer.Transformer):
    """Defines a big Transformer model with shared embeddings as described in
    https://arxiv.org/abs/1706.03762.
    """

    def __init__(self):
        super().__init__(
            num_units=1024,
            num_heads=16,
            ffn_inner_dim=4096,
            share_embeddings=sequence_to_sequence.EmbeddingsSharingLevel.ALL,
        )


@register_model_in_catalog
class TransformerBigRelative(transformer.Transformer):
    """Defines a big Transformer model using relative position representations as
    described in https://arxiv.org/abs/1803.02155.
    """

    def __init__(self):
        super().__init__(
            num_units=1024,
            num_heads=16,
            ffn_inner_dim=4096,
            position_encoder_class=None,
            maximum_relative_position=20,
        )


@register_model_in_catalog
class TransformerTiny(transformer.Transformer):
    """Defines a tiny Transformer model."""

    def __init__(self):
        super().__init__(
            num_layers=2,
            num_units=64,
            num_heads=2,
            ffn_inner_dim=64,
        )


@register_model_in_catalog
class GPT2Small(language_model.LanguageModel):
    """GPT-2 language model (small version) as described in:

    https://d4mucfpksywv.cloudfront.net/better-language-models/language-models.pdf
    """

    def __init__(self):
        super().__init__(
            decoder=decoders.SelfAttentionDecoder(
                num_layers=12,
                num_units=768,
                num_heads=12,
                ffn_inner_dim=3072,
                ffn_activation=layers.gelu,
                position_encoder_class=lambda: layers.PositionEmbedder(
                    maximum_position=1024
                ),
                num_sources=0,
            ),
            embedding_size=768,
        )

    def auto_config(self, num_replicas=1):
        config = super().auto_config(num_replicas=num_replicas)
        return config_util.merge_config(
            config,
            {
                "params": {
                    "average_loss_in_time": True,
                    "optimizer": "Adam",
                    "learning_rate": 2.5e-4,
                    "decay_type": "CosineAnnealing",
                    "decay_params": {
                        "max_step": 1000000,
                        "warmup_steps": 2000,
                    },
                },
                "train": {
                    # Below options are from GPT-1.
                    "batch_size": 64,
                    "maximum_features_length": 512,
                },
            },
        )



"""Define convolution-based encoders."""

import tensorflow as tf

from opennmt.encoders.encoder import Encoder
from opennmt.layers import common, position


class ConvEncoder(Encoder):
    """An encoder that applies a convolution over the input sequence
    as described in https://arxiv.org/abs/1611.02344.
    """

    def __init__(
        self,
        num_layers_a,
        num_layers_c,
        num_units,
        kernel_size=3,
        dropout=0.3,
        position_encoder_class=position.PositionEmbedder,
    ):
        """Initializes the parameters of the encoder.

        Args:
          num_layers_a: The number of layers in CNN-a.
          num_layers_c: The number of layers in CNN-c.
          num_units: The number of output filters.
          kernel_size: The kernel size.
          dropout: The probability to drop units from the inputs.
          position_encoder_class: The :class:`opennmt.layers.PositionEncoder`
            class to use for position encoding (or a callable that returns an
            instance).
        """
        super().__init__()
        self.dropout = dropout
        self.position_encoder = None
        if position_encoder_class is not None:
            self.position_encoder = position_encoder_class()
        self.cnn_a = [
            tf.keras.layers.Conv1D(num_units, kernel_size, padding="same")
            for _ in range(num_layers_a)
        ]
        self.cnn_c = [
            tf.keras.layers.Conv1D(num_units, kernel_size, padding="same")
            for _ in range(num_layers_c)
        ]

    def call(self, inputs, sequence_length=None, training=None):
        if self.position_encoder is not None:
            inputs = self.position_encoder(inputs)
        inputs = common.dropout(inputs, self.dropout, training=training)

        cnn_a = _cnn_stack(self.cnn_a, inputs)
        cnn_c = _cnn_stack(self.cnn_c, inputs)

        outputs = cnn_a
        state = tf.reduce_mean(cnn_c, axis=1)
        return (outputs, state, sequence_length)


def _cnn_stack(layers, inputs):
    next_input = inputs

    for i, layer in enumerate(layers):
        outputs = layer(next_input)
        # Add residual connections past the first layer.
        if i > 0:
            outputs += next_input
        next_input = tf.tanh(outputs)

    return next_input



"""Define self-attention decoder."""

import tensorflow as tf

from opennmt.decoders import decoder
from opennmt.layers import common, transformer
from opennmt.layers.position import SinusoidalPositionEncoder


class SelfAttentionDecoder(decoder.Decoder):
    """Encoder using self-attention as described in
    https://arxiv.org/abs/1706.03762.
    """

    def __init__(
        self,
        num_layers,
        num_units=512,
        num_heads=8,
        ffn_inner_dim=2048,
        dropout=0.1,
        attention_dropout=0.1,
        ffn_dropout=0.1,
        ffn_activation=tf.nn.relu,
        position_encoder_class=SinusoidalPositionEncoder,
        num_sources=1,
        maximum_relative_position=None,
        attention_reduction=transformer.MultiHeadAttentionReduction.FIRST_HEAD_LAST_LAYER,
        pre_norm=True,
        **kwargs
    ):
        """Initializes the parameters of the decoder.

        Args:
          num_layers: The number of layers.
          num_units: The number of hidden units.
          num_heads: The number of heads in the multi-head attention.
          ffn_inner_dim: The number of units of the inner linear transformation
            in the feed forward layer.
          dropout: The probability to drop units from the outputs.
          attention_dropout: The probability to drop units from the attention.
          ffn_dropout: The probability to drop units from the activation output in
            the feed forward layer.
          ffn_activation: The activation function to apply between the two linear
            transformations of the feed forward layer.
          position_encoder_class: The :class:`opennmt.layers.PositionEncoder`
            class to use for position encoding (or a callable that returns an
            instance).
          num_sources: The number of source contexts expected by this decoder.
          maximum_relative_position: Maximum relative position representation
            (from https://arxiv.org/abs/1803.02155).
          attention_reduction: A :class:`opennmt.layers.MultiHeadAttentionReduction`
            value to specify how to reduce multi-head attention matrices.
          pre_norm: If ``True``, layer normalization is applied before each
            sub-layer. Otherwise it is applied after.
          **kwargs: Additional layer arguments.
        """
        super().__init__(num_sources=num_sources, **kwargs)
        self.num_units = num_units
        self.num_heads = num_heads
        self.dropout = dropout
        self.attention_reduction = attention_reduction
        self.position_encoder = None
        if position_encoder_class is not None:
            self.position_encoder = position_encoder_class()
        self.layer_norm = common.LayerNorm() if pre_norm else None
        self.layers = [
            transformer.SelfAttentionDecoderLayer(
                self.num_units,
                self.num_heads,
                ffn_inner_dim,
                num_sources=num_sources,
                dropout=dropout,
                attention_dropout=attention_dropout,
                ffn_dropout=ffn_dropout,
                ffn_activation=ffn_activation,
                maximum_relative_position=maximum_relative_position,
                pre_norm=pre_norm,
            )
            for i in range(num_layers)
        ]

    @property
    def minimum_sources(self):
        return 0

    @property
    def maximum_sources(self):
        return 1e6  # An arbitrary large number.

    @property
    def support_alignment_history(self):
        return True

    def map_v1_weights(self, weights):
        m = super().map_v1_weights(weights)
        m += self.layer_norm.map_v1_weights(weights["LayerNorm"])
        for i, layer in enumerate(self.layers):
            m += layer.map_v1_weights(weights["layer_%d" % i])
        return m

    def _run(
        self,
        inputs,
        sequence_length=None,
        cache=None,
        memory=None,
        memory_sequence_length=None,
        step=None,
        training=None,
    ):
        # Process inputs.
        inputs *= self.num_units**0.5
        if self.position_encoder is not None:
            inputs = self.position_encoder(
                inputs, position=step + 1 if step is not None else None
            )
        inputs = common.dropout(inputs, self.dropout, training=training)

        # Prepare query mask.
        mask = None
        if step is None:
            maximum_length = tf.shape(inputs)[1]
            if sequence_length is None:
                batch_size = tf.shape(inputs)[0]
                sequence_length = tf.fill([batch_size], maximum_length)
            mask = transformer.future_mask(
                sequence_length, maximum_length=maximum_length
            )

        # Prepare memory mask.
        memory_mask = None
        if memory is not None:
            if not isinstance(memory, (list, tuple)):
                memory = (memory,)
            if memory_sequence_length is not None:
                if not isinstance(memory_sequence_length, (list, tuple)):
                    memory_sequence_length = (memory_sequence_length,)
                memory_mask = [
                    tf.sequence_mask(mem_length, maxlen=tf.shape(mem)[1])
                    for mem, mem_length in zip(memory, memory_sequence_length)
                ]
            else:
                memory_mask = tuple(None for _ in memory)

        # Run each layer.
        new_cache = []
        attention = []
        for i, layer in enumerate(self.layers):
            inputs, layer_cache, layer_attention = layer(
                inputs,
                mask=mask,
                memory=memory,
                memory_mask=memory_mask,
                cache=cache[i] if cache is not None else None,
                training=training,
            )
            attention.append(layer_attention)
            new_cache.append(layer_cache)
        outputs = self.layer_norm(inputs) if self.layer_norm is not None else inputs

        # Convert list of shape num_layers x num_sources to num_sources x num_layers
        attention = list(map(list, zip(*attention)))
        if attention:
            attention = transformer.MultiHeadAttentionReduction.reduce(
                attention[0],  # Get attention to the first source.
                self.attention_reduction,
            )
        else:
            attention = None

        return outputs, new_cache, attention

    def forward(
        self,
        inputs,
        sequence_length=None,
        initial_state=None,
        memory=None,
        memory_sequence_length=None,
        input_fn=None,
        sampling_probability=None,
        training=None,
    ):
        _ = initial_state
        _ = input_fn
        if sampling_probability is not None:
            raise ValueError("Scheduled sampling is not supported by this decoder")
        outputs, state, attention = self._run(
            inputs,
            sequence_length=sequence_length,
            memory=memory,
            memory_sequence_length=memory_sequence_length,
            training=training,
        )
        logits = self.output_layer(outputs)
        return logits, state, attention

    def step(
        self,
        inputs,
        timestep,
        state=None,
        memory=None,
        memory_sequence_length=None,
        training=None,
    ):
        inputs = tf.expand_dims(inputs, 1)
        outputs, state, attention = self._run(
            inputs,
            cache=state,
            memory=memory,
            memory_sequence_length=memory_sequence_length,
            step=timestep,
            training=training,
        )
        outputs = tf.squeeze(outputs, axis=1)
        if attention is not None:
            attention = tf.squeeze(attention, axis=1)
        return outputs, state, attention

    def _get_initial_state(self, batch_size, dtype, initial_state=None):
        # The decoder state contains the keys and values projections of the previous timesteps.
        _ = initial_state
        cache = []
        for _ in self.layers:
            shape = [batch_size, self.num_heads, 0, self.num_units // self.num_heads]
            self_kv = (tf.zeros(shape, dtype=dtype), tf.zeros(shape, dtype=dtype))
            memory_kv = [
                (tf.zeros(shape, dtype=dtype), tf.zeros(shape, dtype=dtype))
                for _ in range(self.num_sources)
            ]
            cache.append(dict(self_kv=self_kv, memory_kv=memory_kv))
        return cache

    def _get_state_reorder_flags(self):
        # We don't need to reorder memory_kv as it is the same for all beams.
        return [
            {
                "self_kv": (True, True),
                "memory_kv": [(False, False) for _ in range(self.num_sources)],
            }
            for _ in self.layers
        ]
